import time
from typing import Dict, List, Union

from fuzzy_search.phrase.phrase_model import PhraseModel
from fuzzy_search.match.phrase_match import PhraseMatch
from fuzzy_search.match.phrase_match import Candidate
from fuzzy_search.match.phrase_match import adjust_match_offsets
from fuzzy_search.match.phrase_match import candidates_to_matches
from fuzzy_search.match.phrase_match import filter_matches_by_overlap
from fuzzy_search.match.skip_match import get_skipmatch_candidates
from fuzzy_search.match.exact_match import index_known_word_offsets
from fuzzy_search.match.exact_match import search_exact_phrases
from fuzzy_search.search.searcher import FuzzySearcher
from fuzzy_search.search.token_searcher import FuzzyTokenSearcher
from fuzzy_search.tokenization.string import score_levenshtein_similarity_ratio
from fuzzy_search.tokenization.token import Tokenizer
from fuzzy_search.tokenization.token import Doc


def get_text_dict(text: Union[str, dict, Doc], ignorecase: bool = False) -> dict:
    """Check that text is in a dictionary with an id property, so that passing a long text
    goes by reference instead of copying the long text string.

    :param text: a text string or text dictionary
    :type text: Union[str, dict]
    :param ignorecase: boolean flag for whether to ignore case
    :type ignorecase: bool
    :return: a text dictionary with an id property
    :rtype: dict
    """
    if isinstance(text, str):
        text = {"text": text, "id": None, 'text_lower': text.lower()}
    elif isinstance(text, Doc):
        text = {'text': text.text, 'id': text.id}
    if "id" not in text:
        text["id"] = None
    return text


def combine_fuzzy_and_exact_matches(filtered_matches: List[PhraseMatch],
                                    exact_matches: List[PhraseMatch],
                                    debug: int = 0):
    combined_matches = [em for em in exact_matches]
    exact_phrases = {(em.offset, em.phrase.phrase_string): em for em in exact_matches}
    if debug > 2:
        print('combine_fuzzy_and_exact_matches - exact_phrases.keys():', exact_phrases.keys())
    for fm in filtered_matches:
        if (fm.offset, fm.phrase.phrase_string) in exact_phrases:
            if debug > 2:
                print('skipping fuzzy match because there is a better exact match:', fm)
            continue
        else:
            combined_matches.append(fm)
    return sorted(combined_matches, key=lambda m: m.offset)


class FuzzyPhraseSearcher(FuzzySearcher):

    def __init__(self, phrase_list: List[any] = None,
                 phrase_model: Union[Dict[str, any], List[Dict[str, any]], PhraseModel] = None,
                 config: Union[None, Dict[str, Union[str, int, float]]] = None,
                 tokenizer: Tokenizer = None,
                 token_searcher: FuzzyTokenSearcher = None):
        """This class represents the basic fuzzy searcher. You can pass a list of phrases or a phrase model and
        configuration dictionary that overrides the default configuration values. The default config dictionary
        is available via `fuzzy_search.default_config`.

        To set e.g. the character ngram_size to 3 and the skip_size to 1 use the following dictionary:

        config = {
            'ngram_size': 3,
            'skip_size': 1
        }

        :param phrase_list: a list of phrases (a list of strings or more complex dictionaries with phrases and variants)
        :type phrase_list: list
        :param phrase_model: a phrase model
        :type phrase_model: PhraseModel
        :param config: a configuration dictionary to override default configuration properties.
        Only the properties in the config dictionaries of updated.
        :type config: dict
        :param tokenizer: a tokenizer instance
        :type tokenizer: Tokenizer
        :param token_searcher: a fuzzy token searcher instance (using the same phrase model and config)
        :type token_searcher: FuzzyTokenSearcher
        """
        super().__init__(phrase_list=phrase_list, phrase_model=phrase_model,
                         config=config, tokenizer=tokenizer)
        if token_searcher is None:
            token_searcher = FuzzyTokenSearcher(phrase_model=self.phrase_model,
                                                config=self.config, tokenizer=self.tokenizer)
        self.token_searcher = token_searcher

    def find_candidates(self, text: dict, use_word_boundaries: bool,
                        include_variants: Union[None, bool] = None,
                        known_word_start_offset: Dict[int, Dict[str, any]] = None,
                        debug: int = 0) -> List[Candidate]:
        """Find candidate fuzzy matches for a given text.

        :param text: the text object to match with phrases
        :type text: dict
        :param use_word_boundaries: use word boundaries in determining match boundaries
        :type use_word_boundaries: bool
        :param include_variants: boolean flag for whether to include phrase variants for finding matches
        :type include_variants: bool
        :param known_word_start_offset: a dictionary of known words and their text offsets based on exact matches
        :type known_word_start_offset: Dict[int, Dict[str, any]]
        :param debug: level to show debug information
        :type debug: int
        :return: a list of candidate matches
        :rtype: List[Candidate]
        """
        skip_matches = self.find_skipgram_matches(text, include_variants=include_variants,
                                                  known_word_start_offset=known_word_start_offset)
        candidates = get_skipmatch_candidates(text, skip_matches, self.skipgram_threshold, self.phrase_model,
                                              max_length_variance=self.max_length_variance,
                                              ignorecase=self.ignorecase, debug=debug)
        if debug > 2:
            print('find_candidates - candidates:')
            for candidate in candidates:
                print(f"\t{candidate.phrase}\t-\t{candidate.match_string}")
        filtered = []
        use_word_boundaries = use_word_boundaries if use_word_boundaries is not None else self.use_word_boundaries
        if not use_word_boundaries:
            filtered = candidates
        else:
            if debug > 1:
                print('find_candidates - start word boundary filtering candidates')
            for candidate in candidates:
                if debug > 3:
                    print()
                    print('find_candidates - candidate:', candidate)
                adjusted_match = adjust_match_offsets(candidate.phrase.phrase_string, candidate.match_string,
                                                      text, candidate.match_start_offset, candidate.match_end_offset,
                                                      self.punctuation, debug=debug)
                if debug > 3:
                    print("find_candidates - adjusted_match:", adjusted_match)
                if not adjusted_match:
                    if debug > 2:
                        print(f"find_candidates - removing candidate '{candidate.phrase}'\t-\t'{candidate.match_string}'")
                    continue
                candidate.match_start_offset = adjusted_match["match_start_offset"]
                candidate.match_end_offset = adjusted_match["match_end_offset"]
                candidate.match_string = adjusted_match["match_string"]
                if debug > 3:
                    print("find_candidates - new match string:", candidate.match_string)
                filtered.append(candidate)
        if debug > 1:
            print('find_candidates - returning word boundary filtered candidates:')
            for candidate in filtered:
                print(f"\t{candidate.phrase.phrase_string}\t-\t{candidate.match_string}")
        return filtered

    def filter_matches_by_distractors(self, matches: List[PhraseMatch]) -> List[PhraseMatch]:
        filtered: List[PhraseMatch] = []
        for match in matches:
            if match.phrase.phrase_string in self.phrase_model.has_distractors:
                for distractor in self.phrase_model.has_distractors[match.phrase.phrase_string]:
                    score = score_levenshtein_similarity_ratio(match.string, distractor)
                    if score > match.levenshtein_similarity:
                        break
                else:
                    filtered.append(match)
            else:
                filtered.append(match)
        return filtered

    def filter_matches_by_threshold(self, matches: List[PhraseMatch]) -> List[PhraseMatch]:
        filtered: List[PhraseMatch] = []
        for match in matches:
            if match.character_overlap < self.char_match_threshold:
                continue
            if match.ngram_overlap < self.ngram_threshold:
                continue
            if match.levenshtein_similarity < self.levenshtein_threshold:
                continue
            filtered.append(match)
        return filtered

    def find_matches(self, text: Union[str, Dict[str, str], Doc],
                     use_word_boundaries: Union[None, bool] = None,
                     allow_overlapping_matches: Union[None, bool] = None,
                     include_variants: Union[None, bool] = None,
                     filter_distractors: Union[None, bool] = None,
                     skip_exact_matching: bool = None,
                     first_best: bool = False,
                     debug: int = 0) -> List[PhraseMatch]:
        """Find all fuzzy matching phrases for a given text. By default, a first pass of exact matching is conducted
        to find exact occurrences of phrases. This is to speed up the fuzzy matching pass

        :param text: the text (string or dictionary with 'text' property) to find fuzzy matching phrases in.
        :type text: Union[str, Dict[str, str]]
        :param use_word_boundaries: use word boundaries in determining match boundaries
        :type use_word_boundaries: bool
        :param allow_overlapping_matches: boolean flag for whether to allow matches to overlap in their text ranges
        :type allow_overlapping_matches: bool
        :param include_variants: boolean flag for whether to include phrase variants for finding matches
        :type include_variants: bool
        :param filter_distractors: boolean flag for whether to remove phrase matches that better match distractors
        :type filter_distractors: bool
        :param skip_exact_matching: boolean flag whether to skip the exact matching step
        :type skip_exact_matching: bool
        :return: a list of phrases matches
        :type first_best: bool
        :return: whether to return only the first match with the best score, or all matches
        :param debug: level to show debug information
        :type debug: int
        :rtype: PhraseMatch
        """
        time_step = make_step_timer()
        if debug > 0:
            print('find_matches - getting text dict')
        if self.phrase_model is None:
            raise ValueError("No phrase model indexed")
        text = get_text_dict(text, ignorecase=self.ignorecase)
        if use_word_boundaries is None:
            use_word_boundaries = self.use_word_boundaries
        if skip_exact_matching is None:
            skip_exact_matching = self.skip_exact_matching
        if not skip_exact_matching:
            if debug > 0:
                time_step()
                print("find_matches - running exact matching")
            exact_matches = self.find_exact_matches(text, use_word_boundaries=use_word_boundaries,
                                                    include_variants=include_variants)
            known_word_start_offset = index_known_word_offsets(exact_matches)
        else:
            if debug > 0:
                time_step()
                print("find_matches - skipping exact matching")
            exact_matches = []
            known_word_start_offset = {}
        if debug > 0:
            time_step()
            print('find_matches - number of exact matches:', len(exact_matches))
        candidates = self.find_candidates(text, use_word_boundaries=use_word_boundaries,
                                          include_variants=include_variants,
                                          known_word_start_offset=known_word_start_offset, debug=debug)
        if debug > 0:
            print('find_matches - number of candidates received from find_candidates:', len(candidates))
        if debug > 0:
            time_step()
        if debug > 2:
            for candidate in candidates:
                print(f"\t{candidate.phrase.phrase_string}\t-\t{candidate.match_string}")
        matches = candidates_to_matches(candidates, text, self.phrase_model, ignorecase=self.ignorecase)
        if debug > 0:
            time_step()
            print('find_macthes - number of matches:', len(matches))
        if debug > 2:
            print('find_macthes - matches:')
            for match in matches:
                print(f"\tmatch phrase_string: {match.phrase.phrase_string}\tmatch_string: {match.string}")
        filtered_matches = self.filter_matches_by_threshold(matches)
        if filter_distractors is None:
            filter_distractors = self.filter_distractors
        if filter_distractors:
            filtered_matches = self.filter_matches_by_distractors(filtered_matches)
            if debug > 1:
                print('find_macthes - number of matches after filtering by distractors:', len(matches))
        if allow_overlapping_matches is None:
            allow_overlapping_matches = self.allow_overlapping_matches
        if debug > 1:
            print(len(filtered_matches), len(exact_matches))
            print([fm.levenshtein_similarity for fm in filtered_matches])
            print([em.levenshtein_similarity for em in exact_matches])
            for fm in filtered_matches:
                print(fm)
        filtered_matches = combine_fuzzy_and_exact_matches(filtered_matches, exact_matches)
        if debug > 1:
            print('find_macthes - number of matches after combining fuzzy and exact matches:', len(filtered_matches))
        filtered_matches = self.filter_matches_by_offset_threshold(filtered_matches, debug=debug)
        if debug > 1:
            print('find_macthes - number of matches after filtering by offset threshold:', len(filtered_matches))
        if not allow_overlapping_matches:
            filtered_matches = filter_matches_by_overlap(filtered_matches, first_best=first_best, debug=debug)
            if debug > 1:
                print('find_macthes - number of matches after filtering by overlap:', len(filtered_matches))
        # print(exact_matches)
        if debug > 0:
            time_step()
            print('find_matches - filtered_matches:', filtered_matches)
        return sorted(filtered_matches, key=lambda x: (x.text_id, x.offset, x.offset + len(x.string)))

    def find_exact_matches(self, text: Union[str, Dict[str, str]],
                           use_word_boundaries: Union[None, bool] = None,
                           include_variants: Union[None, bool] = None,
                           debug: int = 0) -> List[PhraseMatch]:
        """Find all fuzzy matching phrases for a given text.

        :param text: the text (string or dictionary with 'text' property) to find fuzzy matching phrases in.
        :type text: Union[str, Dict[str, str]]
        :param use_word_boundaries: use word boundaries in determining match boundaries
        :type use_word_boundaries: Union[None, bool]
        :param include_variants: boolean flag for whether to include phrase variants for finding matches
        :type include_variants: Union[None, bool]
        :param debug: level to show debug information
        :type debug: int
        :return: a list of phrases matches
        :rtype: PhraseMatch
        """
        exact_matches: List[PhraseMatch] = []
        text = get_text_dict(text, ignorecase=self.ignorecase)
        if use_word_boundaries is None:
            use_word_boundaries = self.use_word_boundaries
        if include_variants is None:
            include_variants = self.include_variants
        if debug > 0:
            print('find_exact_matches - use_word_boundaries:', use_word_boundaries)
            print('find_exact_matches - include_variants:', include_variants)
        for exact_match in search_exact_phrases(self.phrase_model, text, use_word_boundaries=use_word_boundaries,
                                                include_variants=include_variants, debug=debug):
            exact_matches.append(exact_match)
        return exact_matches


def make_step_timer():

    first_step = time.time()
    prev_step = first_step

    def time_step():
        nonlocal prev_step
        curr_step = time.time()
        took = curr_step - prev_step
        prev_step = curr_step
        print(f'\tstep took {took: >.2f} seconds, total: {curr_step - first_step: >.2f}')
        return took

    return time_step




from unittest import TestCase

from fuzzy_search.tokenization.token import Tokenizer
from fuzzy_search.tokenization.vocabulary import Vocabulary


class TestVocabulary(TestCase):

    def setUp(self) -> None:
        self.words = ['This', 'is', 'a', 'list', 'of', 'words']
        text = ' '.join(self.words)
        tokenizer = Tokenizer()
        self.doc = tokenizer.tokenize_doc(text)

    def test_vocabulary_can_be_initialised_empty(self):
        vocab = Vocabulary()
        self.assertEqual(0, len(vocab.term_id))

    def test_vocabulary_can_be_initialised_with_word_list(self):
        vocab = Vocabulary(self.words)
        self.assertEqual(len(self.words), len(vocab.term_id))

    def test_vocabulary_can_be_initialised_with_token_list(self):
        vocab = Vocabulary(self.doc.tokens)
        self.assertEqual(len(self.words), len(vocab.term_id))

    def test_vocabulary_represents_words_by_integer(self):
        vocab = Vocabulary(self.words)
        self.assertEqual(True, isinstance(vocab.term_id[self.words[0]], int))

    def test_vocabulary_can_check_that_word_is_included(self):
        vocab = Vocabulary(self.words)
        self.assertEqual(True, vocab.has_term('This'))

    def test_vocabulary_can_check_that_token_is_included(self):
        vocab = Vocabulary(self.doc.tokens)
        self.assertEqual(True, vocab.has_term(self.doc.tokens[0]))

    def test_vocabulary_can_be_case_insensitive(self):
        vocab = Vocabulary(self.words, ignorecase=True)
        self.assertEqual(True, vocab.has_term('this'))

    def test_vocabulary_can_be_case_sensitive(self):
        vocab = Vocabulary(self.words, ignorecase=False)
        self.assertEqual(True, vocab.has_term('This'))
        self.assertEqual(False, vocab.has_term('this'))



from unittest import TestCase

from fuzzy_search.phrase.phrase import Phrase
from fuzzy_search.tokenization.string import SkipGram
from fuzzy_search.match.skip_match import SkipMatches


class TestSkipMatches(TestCase):

    def test_skip_matches_registers_match(self):
        skip_matches = SkipMatches(2, 2)
        phrase = Phrase('test')
        skipgram = SkipGram('ts', 0, len(phrase.phrase_string), 3)
        skip_matches.add_skip_match(skipgram, phrase)
        self.assertTrue(phrase in skip_matches.match_set)




import string

default_config = {
    # these thresholds work when there are quite a few OCR errors
    # use higher thresholds for higher quality OCR to avoid
    # spurious matches.
    "char_match_threshold": 0.6,
    "ngram_threshold": 0.5,
    "levenshtein_threshold": 0.6,
    "skipgram_threshold": 0.2,
    # Is upper/lowercase a meaningful signal?
    "ignorecase": False,
    # should matches follow word boundaries?
    "use_word_boundaries": False,
    # for phrases that have variant phrasings
    "include_variants": False,
    # avoid matching with similar but different phrases
    "filter_distractors": False,
    # matching string can be lower/shorter than prhase
    "max_length_variance": 1,
    # higher ngram size allows fewer character differences
    "ngram_size": 2,
    # fewer skips is much faster but less exhaustive
    "skip_size": 2,
    # first check for exact matches to speed up fuzzy search
    "skip_exact_matching": False,
    # allow matches of partially overlapping phrase
    "allow_overlapping_matches": False,
    # the set of symbols to use as punctuation (for word boundaries)
    "punctuation": string.punctuation,
    "debug": False
}


from unittest import TestCase

import fuzzy_search.similarity as similarity


class TestKeywordList(TestCase):

    def setUp(self) -> None:
        self.keywords = ['test', 'of', 'words' 'of', 'different', 'length']

    def test_keyword_list_accepts_list_of_string(self):
        error = None
        try:
            similarity.KeywordList(self.keywords, max_length_diff=2)
        except BaseException as err:
            error = err
        self.assertEqual(None, error)

    def test_keyword_list_does_not_accept_list_of_dicts(self):
        self.assertRaises(ValueError, similarity.KeywordList, [{'a': 1, 'b': 2}], 2)

    def test_keyword_list_indexes_multiple_occurrences_only_once(self):
        kwl = similarity.KeywordList(self.keywords, max_length_diff=2)
        self.assertEqual(1, kwl.len_keys[2].count('of'))

    def test_keyword_list_indexes_by_length(self):
        kwl = similarity.KeywordList(self.keywords, max_length_diff=2)
        for ti, test_kw in enumerate(self.keywords):
            with self.subTest(ti):
                self.assertIn(test_kw, kwl.len_keys[len(test_kw)])

    def test_keyword_pairs_obey_length_restrictions(self):
        max_length_diff = 2
        kwl = similarity.KeywordList(self.keywords, max_length_diff=max_length_diff)
        for pi, pair in enumerate(kwl.iterate_candidate_pairs()):
            with self.subTest(pi):
                kw1, kw2 = pair
                self.assertEqual(max_length_diff, max(max_length_diff, abs(len(kw1) - len(kw2))))



from typing import Generator
from unittest import TestCase

from fuzzy_search.phrase.phrase import text2skipgrams, Phrase
from fuzzy_search.tokenization.token import Tokenizer


class Test(TestCase):

    def test_text2skipgrams_rejects_negative_ngram_size(self):
        error = None
        try:
            next(text2skipgrams("test", ngram_size=-1, skip_size=2))
        except ValueError as err:
            error = err
        self.assertNotEqual(error, None)

    def test_text2skipgrams_rejects_negative_skip_size(self):
        error = None
        try:
            next(text2skipgrams("test", ngram_size=2, skip_size=-1))
        except ValueError as err:
            error = err
        self.assertNotEqual(error, None)

    def test_text2skipgrams_accepts_positive_ngram_size(self):
        skip_iter = text2skipgrams("test", ngram_size=2, skip_size=2)
        self.assertEqual(isinstance(skip_iter, Generator), True)

    def test_skipgrams_have_correct_length(self):
        skipgrams = [skipgram for skipgram in text2skipgrams("test", ngram_size=2, skip_size=2)]
        self.assertEqual(skipgrams[2].length, 4)


class TestFuzzyPhrase(TestCase):

    def test_fuzzy_phrase_accepts_phrase_as_string(self):
        phrase = Phrase("some phrase")
        self.assertEqual(phrase.phrase_string, "some phrase")

    def test_fuzzy_phrase_accepts_phrase_as_dict(self):
        phrase = Phrase({"phrase": "some phrase"})
        self.assertEqual(phrase.phrase_string, "some phrase")

    def test_fuzzy_phrase_can_set_max_start_offset(self):
        phrase = Phrase({"phrase": "some phrase", "max_start_offset": 3})
        self.assertEqual(phrase.max_start_offset, 3)

    def test_fuzzy_phrase_can_check_if_max_start_offset(self):
        phrase = Phrase({"phrase": "some phrase", "max_start_offset": 3})
        self.assertEqual(True, phrase.has_max_start_offset())

    def test_fuzzy_phrase_can_set_max_start_end(self):
        phrase = Phrase({"phrase": "some phrase", "max_start_offset": 3})
        self.assertEqual(phrase.max_start_end, 3 + len("some_phrase"))

    def test_fuzzy_phrase_can_check_if_max_end_offset(self):
        phrase = Phrase({"phrase": "some phrase", "max_end_offset": 3})
        self.assertEqual(True, phrase.has_max_end_offset())

    def test_fuzzy_phrase_cannot_set_negative_max_start_offset(self):
        error = None
        try:
            Phrase({"phrase": "some phrase", "max_start_offset": -3})
        except ValueError as err:
            error = err
        self.assertNotEqual(error, None)

    def test_fuzzy_phrase_accepts_phrase_with_valid_string_label(self):
        phrase = Phrase({"phrase": "some phrase", "label": "some_label"})
        self.assertEqual(phrase.label, "some_label")

    def test_fuzzy_phrase_accepts_phrase_with_valid_list_of_strings_label(self):
        phrase = Phrase({"phrase": "some phrase", "label": ["some_label", "other_label"]})
        self.assertEqual(isinstance(phrase.label, list), True)
        self.assertEqual("other_label" in phrase.label, True)

    def test_fuzzy_phrase_rejects_phrase_with_invalid_label(self):
        error = None
        try:
            Phrase({"phrase": "some phrase", "label": {"complex": "label"}})
        except ValueError as err:
            error = err
        self.assertNotEqual(error, None)

    def test_fuzzy_phrase_can_set_metadata(self):
        phrase = Phrase({"phrase": "some phrase", "metadata": {"lang": "en"}})
        self.assertEqual('en', phrase.metadata['lang'])


class TestPhraseTokens(TestCase):

    def setUp(self) -> None:
        self.phrase = "some phrase"
        self.tokenizer = Tokenizer()
        self.doc = self.tokenizer.tokenize_doc(self.phrase)

    def test_phrase_can_take_tokens(self):
        phrase = Phrase(self.phrase, tokens=self.doc.tokens)
        for doc_token, phrase_token in zip(self.doc.tokens, phrase.tokens):
            with self.subTest(doc_token.char_index):
                self.assertEqual(doc_token, phrase_token)

    def test_phrase_can_take_tokenizer(self):
        phrase = Phrase(self.phrase, tokenizer=self.tokenizer)
        for doc_token, phrase_token in zip(self.doc.tokens, phrase.tokens):
            with self.subTest(doc_token.char_index):
                self.assertEqual(doc_token.char_index, phrase_token.char_index)

    def test_phrase_with_tokens_has_token_index(self):
        phrase = Phrase(self.phrase, tokenizer=self.tokenizer)
        self.assertEqual(True, 0 in phrase.token_index['some'])

    def test_phrase_with_tokens_has_token_multi_index(self):
        phrase = Phrase('multiple occurs multiple times', tokenizer=self.tokenizer)
        self.assertEqual(2, len(phrase.token_index['multiple']))



import re
from typing import Dict, List, Set
from collections import defaultdict

from fuzzy_search.phrase.phrase import Phrase
from fuzzy_search.phrase.phrase_model import PhraseModel
from fuzzy_search.match.phrase_match import PhraseMatch


def index_known_word_offsets(exact_matches: List[PhraseMatch]) -> Dict[int, Dict[str, any]]:
    exact_match_offset: Dict[int, Set[PhraseMatch]] = defaultdict(set)
    known_word_offset: Dict[int, Dict[str, any]] = defaultdict(dict)
    for exact_match in exact_matches:
        match_words = re.split(r"\W+", exact_match.string)
        # print("exact match:", exact_match)
        text_offset = exact_match.offset
        match_offset = 0
        # print("text_offset:", text_offset, "\tmatch_offset:", match_offset)
        for match_word in match_words:
            start = text_offset + match_offset + exact_match.string[match_offset:].index(match_word)
            # print("match_word:", match_word, "\tstart:", start)
            end = start + len(match_word)
            if start not in known_word_offset:
                known_word = {
                    "word": match_word,
                    "start": start,
                    "end": end,
                    "match_phrases": {exact_match.string}
                }
                # print(known_word)
                known_word_offset[start] = known_word
            known_word_offset[start]["match_phrases"].add(exact_match.string)
            match_word_offset = match_offset + exact_match.string[match_offset:].index(match_word)
            # print("text_offset:", text_offset, "\tmatch_offset:", match_offset)
        exact_match_offset[exact_match.offset].add(exact_match)
    return known_word_offset


def search_exact_phrases(phrase_model: PhraseModel, text: Dict[str, str],
                         ignorecase: bool = False, use_word_boundaries: bool = True,
                         include_variants: bool = False, debug: int = 0):
    if use_word_boundaries:
        # print('searching with word boundaries')
        return search_exact_phrases_with_word_boundaries(phrase_model, text, ignorecase=ignorecase,
                                                         include_variants=include_variants, debug=debug)
    else:
        return search_exact_phrases_without_word_boundaries(phrase_model, text, ignorecase=ignorecase,
                                                            include_variants=include_variants, debug=debug)


def add_exact_match_score(match: PhraseMatch) -> PhraseMatch:
    match.character_overlap = 1.0
    match.ngram_overlap = 1.0
    match.levenshtein_similarity = 1.0
    return match


def search_exact_phrases_with_word_boundaries(phrase_model: PhraseModel, text: Dict[str, str],
                                              ignorecase: bool = False, include_variants: bool = False,
                                              debug: int = 0):
    for word in re.finditer(r"\w+", text["text"]):
        if debug > 0:
            print('search_exact_phrases_with_word_boundaries - word:', word)
            print('search_exact_phrases_with_word_boundaries - word in phrase_model:', word.group(0) not in phrase_model.word_in_phrase)
        if word.group(0) not in phrase_model.word_in_phrase:
            continue
        if debug > 0:
            print("\tsearch_exact_phrases_with_word_boundaries - matching word:", word)
        for phrase_string in phrase_model.first_word_in_phrase[word.group(0)]:
            phrase_word_offset = phrase_model.first_word_in_phrase[word.group(0)][phrase_string]
            phrase_start = word.start() - phrase_word_offset
            phrase_end = phrase_start + len(phrase_string)
            if debug > 0:
                print('\tsearch_exact_phrases_with_word_boundaries - start, end, string:', phrase_start, phrase_end,
                      phrase_string)
                print('\tsearch_exact_phrases_with_word_boundaries - text start, end:',
                      text["text"][phrase_start:phrase_end])
                print('\tsearch_exact_phrases_with_word_boundaries - same?:',
                      text["text"][phrase_start:phrase_end] == phrase_string)

            if text["text"][phrase_start:phrase_end] == phrase_string:
                if phrase_start > 0 and re.match(r'\w', text["text"][phrase_start - 1]):
                    if debug > 0:
                        print('\tsearch_exact_phrases_with_word_boundaries - match word is not at start word boundary')
                    continue
                if phrase_end < len(text['text']) - 1 and re.match(r'\w', text['text'][phrase_end]):
                    if debug > 0:
                        print('\tsearch_exact_phrases_with_word_boundaries - match word is not at end word boundary')
                    continue

                if debug > 0:
                    print('\tsearch_exact_phrases_with_word_boundaries - phrase_string, phrase_type:', phrase_string,
                      phrase_model.phrase_type[phrase_string])
                if "phrase" in phrase_model.phrase_type[phrase_string]:
                    if debug > 0:
                        print('\tsearch_exact_phrases_with_word_boundaries - match word is phrase')
                    phrase = phrase_model.phrase_index[phrase_string]
                    match = PhraseMatch(phrase, phrase, phrase_string, phrase_start, text_id=text["id"],
                                        ignorecase=ignorecase)
                    yield add_exact_match_score(match)
                    if debug > 0:
                        print("\tsearch_exact_phrases_with_word_boundaries - the matching phrase:", phrase)
                elif "variant" in phrase_model.phrase_type[phrase_string] and include_variants:
                    if debug > 0:
                        print('\tsearch_exact_phrases_with_word_boundaries - match word is variant')
                    variant_phrase = phrase_model.variant_index[phrase_string]
                    main_phrase_string = phrase_model.is_variant_of[phrase_string]
                    main_phrase = phrase_model.phrase_index[main_phrase_string]
                    match = PhraseMatch(main_phrase, variant_phrase, phrase_string, phrase_start,
                                        text_id=text["id"], ignorecase=ignorecase)
                    yield add_exact_match_score(match)


def search_exact_phrases_without_word_boundaries(phrase_model: PhraseModel, text: Dict[str, str],
                                                 ignorecase: bool = False,
                                                 include_variants: bool = False, debug: int = 0):
    for phrase_string in phrase_model.phrase_index:
        phrase = phrase_model.phrase_index[phrase_string]
        if debug > 0:
            print('search_exact_phrases_without_word_boundaries - phrase_string:', phrase_string)
        for match in re.finditer(phrase.exact_string, text["text"]):
            if debug > 0:
                print('search_exact_phrases_without_word_boundaries - match:', match)
            phrase = phrase_model.phrase_index[phrase_string]
            match = PhraseMatch(phrase, phrase, phrase_string, match.start(), text_id=text["id"],
                                ignorecase=ignorecase)
            yield add_exact_match_score(match)
    if include_variants:
        for phrase_string in phrase_model.variant_index:
            if debug > 0:
                print('search_exact_phrases_without_word_boundaries - phrase_string:', phrase_string)
            variant_phrase = phrase_model.variant_index[phrase_string]
            for match in re.finditer(variant_phrase.exact_string, text["text"]):
                if debug > 0:
                    print('search_exact_phrases_without_word_boundaries - match:', match)
                variant_phrase = phrase_model.variant_index[phrase_string]
                main_phrase_string = phrase_model.is_variant_of[phrase_string]
                main_phrase = phrase_model.phrase_index[main_phrase_string]
                match = PhraseMatch(main_phrase, variant_phrase, phrase_string, match.start(),
                                    text_id=text["id"], ignorecase=ignorecase)
                yield add_exact_match_score(match)


def search_exact(phrase: Phrase, text: Dict[str, str], ignorecase: bool = False, use_word_boundaries: bool = True):
    search_string = phrase.extact_word_boundary_string if use_word_boundaries else phrase.exact_string
    if ignorecase:
        return re.finditer(search_string, text["text"], flags=re.IGNORECASE)
    else:
        return re.finditer(search_string, text["text"])


def get_known_word_offsets(match_ranges: List[Dict[str, any]], text_doc: Dict[str, str]) -> Dict[int, dict]:
    known_word_offset = {}
    offset = match_ranges[0]["s"]
    for match_range in match_ranges:
        print(match_range)
        print("offset:", offset)
        match_text = text_doc["text"][match_range["s"]:match_range["e"]]
        match_words = re.split(r"\W+", match_text)
        for match_word in match_words:
            start = offset + text_doc["text"][offset:].index(match_word)
            end = start + len(match_word)
            known_word = {
                "match_word": match_word,
                "start": start,
                "end": end,
                "match_phrases": set([phrase for phrase in match_range["phrases"]])
            }
            match_word_offset = offset + text_doc["text"][offset:].index(match_word)
            offset = match_word_offset + len(match_word)
            print(known_word)
            known_word_offset[known_word["start"]] = known_word
    return known_word_offset


def get_exact_match_ranges(exact_matches: List[PhraseMatch]) -> List[dict]:
    sorted_matches = sorted(exact_matches, key=lambda m: m.offset)
    match_ranges = []
    if len(exact_matches) == 0:
        return []
    first_match = exact_matches[0]
    match_range = {"s": first_match.offset, "e": first_match.end, "phrases": {first_match.phrase.phrase_string}}
    for mi, match in enumerate(sorted_matches[1:]):
        if match.offset > match_range["e"]:
            match_ranges.append(match_range)
            match_range = {"s": match.offset, "e": match.end, "phrases": set()}
        match_range["e"] = max(match_range["e"], match.end)
        match_range["phrases"].add(match.phrase.phrase_string)
        if match.phrase.phrase_string != match.variant.phrase_string:
            match_range["phrases"].add(match.variant.phrase_string)
    match_ranges.append(match_range)
    return match_ranges



import copy
import re
from collections import defaultdict
from typing import Callable, Dict, List, Set, Union


class Annotation:
    """
    Represents an annotation in a document.

    Attributes:
        tag_type (str): The type of the tag.
        text (str): The text of the annotation.
        offset (int): The starting index of the annotation.
        end (int): The ending index of the annotation (offset + length of the text).
        doc_id (str, optional): The ID of the document the annotation belongs to.
    """

    def __init__(self, tag_type: str, text: str, offset: int, doc_id: str = None):
        """
        Initializes the Annotation instance.

        Args:
            tag_type (str): The type of the tag.
            text (str): The text of the annotation.
            offset (int): The starting index of the annotation.
            doc_id (str, optional): The ID of the document the annotation belongs to.
        """
        self.tag_type = tag_type
        self.text = text
        self.offset = offset
        self.end = offset + len(text)
        self.doc_id = doc_id


class Tag:
    """
    Represents a tag in a document, containing both the tag's type and text.

    Attributes:
        type (str): The type of the tag.
        text (str): The text content of the tag.
        offset (int): The starting index of the tag in the document.
        tag_string (str): The full tag string, including opening and closing tags.
        end (int): The ending index of the tag (offset + length of the tag string).
        doc_id (str, optional): The ID of the document the tag belongs to.
    """

    def __init__(self, tag_type: str, text: str, offset: int, doc_id: str = None):
        """
        Initializes the Tag instance.

        Args:
            tag_type (str): The type of the tag.
            text (str): The text content of the tag.
            offset (int): The starting index of the tag.
            doc_id (str, optional): The ID of the document the tag belongs to.
        """
        self.type = tag_type
        self.text = text
        self.offset = offset
        self.tag_string = f"<{tag_type}>{text}</{tag_type}>"
        self.end = offset + len(self.tag_string)
        self.doc_id = doc_id


class Token:
    """
    Represents a token in a document.

    Attributes:
        string (str): The original string of the token.
        index (int): The index of the token in the document.
        char_index (int): The character index of the token in the document.
        doc_id (str, optional): The ID of the document the token belongs to.
        metadata (dict, optional): Additional metadata associated with the token.
        normalised_string (str): The normalized version of the token.
        label (set): A set of labels assigned to the token.
    """

    def __init__(self, string: str, index: int, char_index: int, doc_id: str = None,
                 normalised_string: str = None, label: Union[str, Set[str]] = None,
                 metadata: Dict[str, any] = None):
        """
        Initializes the Token instance.

        Args:
            string (str): The original string of the token.
            index (int): The index of the token in the document.
            char_index (int): The character index of the token in the document.
            doc_id (str, optional): The ID of the document the token belongs to.
            normalised_string (str, optional): The normalized version of the token.
            label (str, set, list, optional): Labels associated with the token.
            metadata (dict, optional): Additional metadata associated with the token.
        """
        self.string = string
        self.index = index
        self.char_index = char_index
        self.doc_id = doc_id
        self.metadata = metadata if metadata else {}
        self.normalised_string = normalised_string if normalised_string else string
        if label is None:
            label = set()
        elif isinstance(label, str):
            label = {label}
        elif isinstance(label, list):
            label = set(label)
        elif isinstance(label, set) is False:
            raise TypeError(f'token label must be of type "str", "set" or "list", not {type(label)}')
        self.label = label

    def __repr__(self):
        """
        Returns a string representation of the Token.

        Returns:
            str: A string representation of the token's normalized string.
        """
        return f"'{self.normalised_string}'"

    def __len__(self):
        """
        Returns the length of the token.

        Returns:
            int: The length of the normalized string of the token.
        """
        return len(self.normalised_string)

    def lower(self):
        """
        Converts the normalized string of the token to lowercase.
        """
        self.normalised_string = self.normalised_string.lower()

    @property
    def i(self):
        """
        Gets the index of the token.

        Returns:
            int: The index of the token.
        """
        return self.index

    @property
    def t(self):
        """
        Gets the original string of the token.

        Returns:
            str: The original string of the token.
        """
        return self.string

    @property
    def n(self):
        """
        Gets the normalized string of the token.

        Returns:
            str: The normalized string of the token.
        """
        return self.normalised_string

    @property
    def l(self):
        """
        Gets the labels associated with the token.

        Returns:
            set: A set of labels associated with the token.
        """
        return self.label

    def has_label(self, label: str):
        """
        Checks if the token has a specific label.

        Args:
            label (str): The label to check for.

        Returns:
            bool: True if the token has the label, False otherwise.
        """
        return label in self.label

    def update(self, normalised_string: str) -> 'Token':
        """
        Creates a new token with an updated normalized string.

        Args:
            normalised_string (str): The updated normalized string.

        Returns:
            Token: A new token with the updated normalized string.
        """
        return Token(string=self.t, index=self.i, char_index=self.char_index,
                     normalised_string=normalised_string,
                     metadata=copy.deepcopy(self.metadata))


class Doc:
    """
    Represents a document containing a list of tokens.

    Attributes:
        text (str): The text of the document.
        id (str): The ID of the document.
        tokens (list): A list of Token objects representing the document's tokens.
        token_orig_set (dict): A dictionary mapping original token strings to lists of tokens.
        token_norm_set (dict): A dictionary mapping normalized token strings to lists of tokens.
        label_token_index (defaultdict): A dictionary mapping labels to sets of tokens.
        metadata (dict): Metadata associated with the document.
        annotations (list): A list of annotations associated with the document.
    """

    def __init__(self, text: str, doc_id: str, tokens: List[Token], metadata: Dict[str, any] = None):
        """
        Initializes a Document instance.

        Args:
            text (str): The text content of the document.
            doc_id (str): The ID of the document.
            tokens (List[Token]): A list of Token objects representing the document's tokens.
            metadata (dict, optional): Metadata associated with the document.
        """
        self.text = text
        self.id = doc_id
        self.tokens = tokens
        self.token_orig_set = {}
        self.token_norm_set = {}
        self.label_token_index = defaultdict(set)
        for token in tokens:
            for label in token.label:
                self.label_token_index[label].add(token)
        self.metadata = metadata if metadata else {}
        self.annotations: List[Annotation] = []
        for token in tokens:
            if token.t not in self.token_orig_set:
                self.token_orig_set[token.t] = [token]
            else:
                self.token_orig_set[token.t].append(token)
            if token.n not in self.token_norm_set:
                self.token_norm_set[token.n] = [token]
            else:
                self.token_norm_set[token.n].append(token)

    def __repr__(self):
        """
        Returns a string representation of the document.

        Returns:
            str: A string representation of the document, including its ID, metadata, and tokens.
        """
        return f"Doc(id='{self.id}', metadata={self.metadata}, text=\"{self.text}\", tokens={self.tokens}"

    def __len__(self):
        """
        Returns the number of normalized tokens in the document.

        Returns:
            int: The number of normalized tokens in the document.
        """
        return len(self.normalized)

    def __getitem__(self, item):
        """
        Retrieves a token by index.

        Args:
            item (int): The index of the token to retrieve.

        Returns:
            Token: The token at the specified index.
        """
        return self.tokens[item]

    def __iter__(self):
        """
        Iterates over the tokens in the document.

        Yields:
            Token: Each token in the document.
        """
        for token in self.tokens:
            yield token

    def _has_original_token(self, token: str) -> bool:
        """
        Checks if the token exists in the original set of tokens.

        Args:
            token (str): The original token string.

        Returns:
            bool: True if the token exists in the original set, False otherwise.
        """
        return token in self.token_orig_set

    def _has_normalised_token(self, token: str) -> bool:
        """
        Checks if the token exists in the normalized set of tokens.

        Args:
            token (str): The normalized token string.

        Returns:
            bool: True if the token exists in the normalized set, False otherwise.
        """
        return token in self.token_norm_set

    def has_token(self, token: Union[str, Token]) -> bool:
        """
        Checks if the document contains a specific token.

        Args:
            token (Union[str, Token]): A token string or a Token object.

        Returns:
            bool: True if the token exists in the document, False otherwise.
        """
        if isinstance(token, str):
            token = self.get_token(token)
            return token is not None
        else:
            return self._has_original_token(token.t) or self._has_normalised_token(token.n)

    def get_token(self, token_string: str) -> Union[Token, List[Token]]:
        """
        Retrieves a token (or tokens) from the document based on the token string.

        Args:
            token_string (str): The token string to look up.

        Returns:
            Union[Token, List[Token]]: A Token object or a list of Token objects that match the token string.
        """
        if token_string in self.token_orig_set:
            token = self.token_orig_set[token_string]
        elif token_string in self.token_norm_set:
            token = self.token_norm_set[token_string]
        else:
            token = None
        if isinstance(token, List) and len(token) == 1:
            token = token[0]
        return token

    def add_annotations(self, annotations: List[Annotation], replace: bool = False):
        """
        Adds annotations to the document.

        Args:
            annotations (List[Annotation]): A list of Annotation objects to add.
            replace (bool): whether to replace existing annotation or add (default is False).
        """
        self.annotations = annotations

    @property
    def original(self):
        """
        Retrieves a list of original tokens in the document.

        Returns:
            List[str]: A list of the original tokens.
        """
        return [token.t for token in self.tokens]

    @property
    def normalized(self):
        """
        Retrieves a list of normalized tokens in the document.

        Returns:
            List[str]: A list of the normalized tokens.
        """
        return [token.n for token in self.tokens]


class Tokenizer:
    """
    A base class for tokenizing a document into tokens.

    Attributes:
        ignorecase (bool): Flag indicating whether to ignore case when tokenizing.
        include_boundary_tokens (bool): Flag indicating whether to include boundary tokens.
        remove_punctuation (bool): Flag indicating whether to remove punctuation from tokens.
    """

    def __init__(self, ignorecase: bool = False, include_boundary_tokens: bool = False,
                 remove_punctuation: bool = False):
        """
        Initializes the Tokenizer instance.

        Args:
            ignorecase (bool, optional): Whether to ignore case when tokenizing. Defaults to False.
            include_boundary_tokens (bool, optional): Whether to include boundary tokens. Defaults to False.
            remove_punctuation (bool, optional): Whether to remove punctuation. Defaults to False.
        """
        self.ignorecase = ignorecase
        self.include_boundary_tokens = include_boundary_tokens
        self.remove_punctuation = remove_punctuation

    def _tokenize(self, text: str) -> List[str]:
        """
        Tokenizes the input text into a list of strings.

        Args:
            text (str): The text to tokenize.

        Returns:
            List[str]: A list of tokens.
        """
        non_whitespace_tokens = [token for token in text.strip().split() if token != '']
        tokens = []
        for nw_token in non_whitespace_tokens:
            if self.remove_punctuation:
                sub_tokens = re.split(r'\W+', nw_token)
            else:
                sub_tokens = re.split(r'(\W+)', nw_token)
            tokens.extend([token for token in sub_tokens if token != ''])
        return tokens

    def tokenize(self, doc_text: str, doc_id: str = None) -> List[Token]:
        """
        Tokenizes the input document text and returns a Doc object.

        Args:
            doc_text (str): The text of the document to tokenize.
            doc_id (str, optional): The ID of the document. Defaults to None.

        Returns:
            Doc: A Doc object containing the tokenized text.
        """
        dummy_text = f"<START> {doc_text} <END>" if self.include_boundary_tokens else doc_text
        token_strings = self._tokenize(doc_text)
        if self.include_boundary_tokens:
            token_strings = ['<START>'] + token_strings + ['<END>']
        tokens = []
        prefix_text = ''
        for ti, token_string in enumerate(token_strings):
            char_index = dummy_text.index(token_string) + len(prefix_text)
            prefix_text += dummy_text[:dummy_text.index(token_string) + len(token_string)]
            dummy_text = dummy_text[dummy_text.index(token_string)+len(token_string):]
            token = Token(token_string, index=ti, char_index=char_index, doc_id=doc_id)
            if self.ignorecase:
                if not self.include_boundary_tokens or (ti != 0 and ti != len(token_strings) - 1):
                    token.lower()
            tokens.append(token)
        return tokens

    def tokenize_doc(self, doc_text: str, doc_id: str = None):
        tokens = self.tokenize(doc_text, doc_id)
        return Doc(text=doc_text, doc_id=doc_id, tokens=tokens)


class RegExTokenizer(Tokenizer):
    """
    A tokenizer that splits text into tokens using a regular expression pattern.

    Attributes:
        split_pattern (str): The regular expression pattern used to split the text into tokens.
    """

    def __init__(self, ignorecase: bool = False, include_boundary_tokens: bool = False,
                 split_pattern: str = r'\b'):
        """
        Initializes the RegExTokenizer instance.

        Args:
            ignorecase (bool, optional): Whether to ignore case when tokenizing. Defaults to False.
            include_boundary_tokens (bool, optional): Whether to include boundary tokens. Defaults to False.
            split_pattern (str, optional): The regular expression pattern used to split the text. Defaults to r'\b'.
        """
        super().__init__(ignorecase=ignorecase, include_boundary_tokens=include_boundary_tokens)
        self.split_pattern = re.compile(split_pattern)

    def _tokenize(self, text: str) -> List[str]:
        """
        Tokenizes the input text into a list of strings using the specified regular expression pattern.

        Args:
            text (str): The text to tokenize.

        Returns:
            List[str]: A list of tokens obtained by splitting the input text.
        """
        # Split the text using the regular expression pattern
        tokens = [token.strip() for token in re.split(self.split_pattern, text) if token.strip() != '']
        non_whitespace_tokens = []
        for token in tokens:
            # Split multi-word tokens (e.g., if there is space between parts)
            split_tokens = token.split(' ')
            non_whitespace_tokens.extend([token for token in split_tokens if token != ''])
        return non_whitespace_tokens


class CustomTokenizer(Tokenizer):
    """
    A tokenizer that uses a custom tokenizer function provided by the user.

    Attributes:
        tokenizer_func (Callable): A user-defined function for tokenizing text.
    """

    def __init__(self, tokenizer_func: Callable, **kwargs):
        """
        Initializes the CustomTokenizer instance.

        Args:
            tokenizer_func (Callable): The custom function to use for tokenizing the text.
            **kwargs: Additional arguments passed to the parent Tokenizer class.
        """
        super().__init__(**kwargs)
        self.tokenizer_func = tokenizer_func

    def _tokenize(self, text: str) -> List[str]:
        """
        Tokenizes the input text using the custom tokenizer function.

        Args:
            text (str): The text to tokenize.

        Returns:
            List[str]: A list of tokens generated by the custom tokenizer function.
        """
        return self.tokenizer_func(text)


def update_token(token: Token, new_normalised: str) -> Token:
    """
    Creates a new Token instance by updating the normalization string of an existing token.

    Args:
        token (Token): The original Token object to be updated.
        new_normalised (str): The new normalized string for the token.

    Returns:
        Token: A new Token object with the updated normalization string.
    """
    return Token(string=token.t, index=token.i, char_index=token.char_index,
                 normalised_string=new_normalised,
                 metadata=copy.deepcopy(token.metadata))


def tokens2string(tokens: List[Token]) -> str:
    string = ''
    for token in tokens:
        if token.char_index > len(string):
            diff = token.char_index - len(string)
            string += ' ' * diff
        string += token.t
    return string



from __future__ import annotations
from collections import Counter
from typing import Dict, List, Union

from fuzzy_search.tokenization.string import SkipGram
from fuzzy_search.phrase.phrase import Phrase


class Candidate:

    def __init__(self, phrase: Phrase, max_length_variance: int = 1,
                 ignorecase: bool = False, debug: int = 0):
        """Create a Candidate instance for a given Phrase object.

        :param phrase: a phrase object
        :type phrase: Phrase
        :param ignorecase: whether to ignore case when matching skip grams
        :type ignorecase: bool
        :param debug: level to show debugging info
        :type debug: int
        """
        self.skipgram_set = set()
        self.skipgram_list: List[SkipGram] = []
        self.skipgram_count = Counter()
        self.phrase = phrase
        self.ignorecase = ignorecase
        self.debug = debug
        if ignorecase:
            self.skipgrams = phrase.skipgrams_lower
            self.skipgram_index = phrase.skipgram_index_lower
            self.skipgram_freq = phrase.skipgram_freq_lower
            self.early_skipgram_index = phrase.early_skipgram_index_lower
            self.late_skipgram_index = phrase.late_skipgram_index_lower
        else:
            self.skipgrams = phrase.skipgrams
            self.skipgram_index = phrase.skipgram_index
            self.skipgram_freq = phrase.skipgram_freq
            self.early_skipgram_index = phrase.early_skipgram_index
            self.late_skipgram_index = phrase.late_skipgram_index
        self.max_length_variance = max_length_variance
        self.max_length = len(self.phrase.phrase_string) + self.max_length_variance
        self.match_start_offset: int = -1
        self.match_end_offset: int = -1
        self.match_string: Union[None, str] = None
        self.skipgram_overlap: float = 0.0

    def __repr__(self):
        return f'Candidate(' + \
               f'phrase: "{self.phrase.phrase_string}", match_string: "{self.match_string}", ' + \
               f'match_start_offset: {self.match_start_offset}, match_end_offset: {self.match_end_offset})'

    def add_skip_match(self, skipgram: SkipGram) -> None:
        """Add a skipgram match between a text and a phrase ot the candidate.

        :param skipgram: a matching skipgram
        :type skipgram: SkipGram
        """
        if len(self.skipgram_list) == 0 and skipgram.string not in self.early_skipgram_index:
            if self.debug > 2:
                print("skipping skipgram as first for candidate:", skipgram.string)
            return None
        self.skipgram_set.add(skipgram.string)
        self.skipgram_list.append(skipgram)
        if self.match_start_offset is None or self.match_start_offset < 0:
            self.match_start_offset = self.get_match_start_offset()
        if skipgram.start_offset + skipgram.length > self.match_end_offset:
            self.match_end_offset = skipgram.start_offset + skipgram.length
        self.skipgram_count.update([skipgram.string])
        if self.debug > 2:
            print("\tadd - skipgram:", skipgram.string, skipgram.start_offset)
            print("\tadd - match length:", self.skip_match_length())
            print("\tadd - list:", [skip.string for skip in self.skipgram_list])
        # check if the candidate string is too long to match the phrase
        # if too long, remove the first skipgrams until the string is short enough
        while self.skip_match_length() > self.max_length and len(self.skipgram_list) > 0:
            self.remove_first_skip()
            self.match_start_offset = self.get_match_start_offset()
            if self.debug > 2:
                print("\tremove - too long - length:", self.skip_match_length())
                print("\tremove - too long - list:", [skip.string for skip in self.skipgram_list])
                print("\tremove - too long - start:", self.match_start_offset, "\tend:", self.match_end_offset)
        while len(self.skipgram_list) > 0 and self.skipgram_list[0].string not in self.early_skipgram_index:
            self.remove_first_skip()
            self.match_start_offset = self.get_match_start_offset()
            if self.debug > 2:
                print("\tremove - no start - length:", self.skip_match_length())
                print("\tremove - no start - list:", [skip.string for skip in self.skipgram_list])
                print("\tremove - no start - start:", self.match_start_offset, "\tend:", self.match_end_offset)

    def shift_start_skip(self) -> bool:
        """Check if there is a later skip that is a better start."""
        if self.skip_match_length() <= len(self.phrase.phrase_string):
            return False
        start_skip = self.skipgram_list[0]
        start_phrase_offset = self.skipgram_index[start_skip.string][0].start_offset
        best_start_phrase_offset = start_phrase_offset
        best_start_index = 0
        best_start_skip = start_skip
        for si, skip in enumerate(self.skipgram_list):
            skip_phrase_offset = self.skipgram_index[skip.string][0].start_offset
            if skip.start_offset - start_skip.start_offset > self.skip_match_length() - len(self.phrase.phrase_string):
                # stop looking for better start when remaining skips result in too short match length
                break
            if skip.start_offset > best_start_skip.start_offset and skip_phrase_offset <= best_start_phrase_offset:
                best_start_index = si
                best_start_skip = skip
                best_start_phrase_offset = skip_phrase_offset
            if skip.string not in self.early_skipgram_index:
                break
        for _ in range(0, best_start_index):
            self.remove_first_skip()
        self.match_start_offset = self.get_match_start_offset()
        return best_start_index > 0

    def remove_first_skip(self) -> None:
        """Remove the first matching skipgram from the list and update the count and set."""
        first_skip = self.skipgram_list.pop(0)
        if self.debug > 3:
            print('\tremove_first_skip - removing first skip')
        # reduce count of first skipgram by 1
        self.skipgram_count[first_skip.string] -= 1
        # if count has dropped to zero, remove skipgram from the set
        if self.skipgram_count[first_skip.string] == 0:
            self.skipgram_set.remove(first_skip.string)

    def skip_match_length(self) -> int:
        """Return the length of the matching string.

        :return: difference between start and end offset
        :rtype: int
        """
        if self.match_start_offset is None:
            return 0
        return self.match_end_offset - self.match_start_offset

    def is_match(self, skipgram_threshold: float):
        """Check if the candidate is a likely match for its corresponding phrase.

        :param skipgram_threshold: the threshold to for how many skipgrams have to match between candidate and phrase
        :type skipgram_threshold: float
        :return: a boolean whether this candidate is a likely match for the phrase
        :rtype: bool
        """
        if len(self.skipgram_list) == 0:
            # there are no matching skipgrams, so no matching string
            if self.debug > 2:
                print('\tis_match - NO MATCH: there are no matching skipgrams, so no matching string')
            return False
        if self.skipgram_list[0].string not in self.early_skipgram_index:
            # the first skipgram of candidate is not in the early skipgrams of phrase
            if self.debug > 2:
                print('\tis_match - NO MATCH: the first skipgram of candidate is not in the early skipgrams of phrase')
            return False
        # length of current skip matches should be no longer than phrase plus max length variance
        # but also no shorter than phrase minus max length variance minus late skips offset from end
        # of phrase.
        if self.skip_match_length() > len(self.phrase.phrase_string) + self.max_length_variance:
            if self.debug > 2:
                print('\tis_match - NO MATCH: skip match length is longer than max length variance of phrase')
            return False
        elif self.skip_match_length() < self.phrase.late_threshold - self.max_length_variance:
            if self.debug > 3:
                print('\tis_match - phrase length:', len(self.phrase.phrase_string))
                print('\tis_match - skip match length:', self.skip_match_length())
                print('\tis_match - max length variance:', self.max_length_variance)
                print('\tis_match - late threshold:', self.phrase.late_threshold)
                print('\tis_match - NO MATCH: skip match length is not within max length variance of phrase')
            return False
        if self.debug > 3:
            print('\tis_match - candidate phrase:', self.phrase.phrase_string)
            print('\tis_match - skip_set_overlap:', self.get_skip_set_overlap())
            print('\tis_match - late threshold:', self.phrase.late_threshold)
        if self.skipgram_list[-1].string not in self.late_skipgram_index:
            if self.debug > 2:
                print("\tis_match - NO MATCH: last skip not in late index")
            # the last skipgram of candidate is not in the late skipgrams of phrase
            return False
        if self.get_skip_set_overlap() < skipgram_threshold:
            if self.debug > 2:
                print("\tis_match - below skipgram threshold:", self.get_skip_set_overlap(), skipgram_threshold)
            return False
        else:
            if self.debug > 2:
                print('\tis_match - MATCH!')
            return True

    def get_skip_set_overlap(self) -> float:
        """Calculate and set skipgram overlap between text and phrase skipgram matches.

        :return: the skipgram overlap
        :rtype: float
        """
        self.skipgram_overlap = len(self.skipgram_set) / len(self.phrase.skipgram_set)
        return self.skipgram_overlap

    def get_skip_count_overlap(self) -> float:
        """Calculate deviation of candidate skipgrams from phrase skipgrams.

        :return: the skipgram overlap (-inf, 1.0]
        :rtype: float
        """
        diff = 0
        total = 0
        for skipgram_string, count in self.skipgram_count.items():
            diff += abs(count - self.skipgram_freq[skipgram_string])
            total += count
        return (total - diff) / self.phrase.num_skipgrams

    def get_match_start_offset(self) -> Union[None, int]:
        """Calculate the start offset of the match.

        :return: the start offset of the match
        :rtype: int
        """
        if len(self.skipgram_list) == 0:
            return None
        first_skip = self.skipgram_list[0]
        first_skip_in_phrase = self.skipgram_index[first_skip.string][0]
        match_start_offset = self.skipgram_list[0].start_offset - first_skip_in_phrase.start_offset
        if self.debug > 3:
            print("\tget_match_start_offset - in match:", first_skip.string, first_skip.start_offset)
            print("\tget_match_start_offset - in phrase:",
                  first_skip_in_phrase.string, first_skip_in_phrase.start_offset)
            print("\tget_match_start_offset - match_start_offset:", match_start_offset)
        return 0 if match_start_offset < 0 else match_start_offset

    def get_match_string(self, text: Dict[str, any]) -> Union[str, None]:
        """Find the matching string of a candidate fuzzy match between a text and a phrase.

        :param text: the text object from which the candidate was derived
        :type text: Dict[str, any]
        :return: the matching string
        :rtype: str
        """
        if self.debug > 2:
            print('\tget_match_string - text:', text)
        if self.match_start_offset == self.match_end_offset:
            raise ValueError('start and end offset cannot be the same')
        if self.match_start_offset > self.match_end_offset:
            raise ValueError('start offset cannot be bigger than end offset')
        return text["text"][self.match_start_offset:self.match_end_offset]
    # TODO: check if first to last offset is too long
    # if not, the match string is probably fine
    # if it is, find the best substring

    def same_candidate(self, other: Candidate):
        """Check if this candidate has the same start and end offsets as another candidate.

        :param other: another candidate for the same phrase and text.
        :type other: Candidate
        :return: this candidate match has the same offsets as the other candidate
        :rtype: bool
        """
        if self.match_start_offset != other.match_start_offset:
            return False
        if self.match_end_offset != other.match_end_offset:
            return False
        else:
            return True




import unittest
from unittest import TestCase

from fuzzy_search.phrase.phrase import Phrase
from fuzzy_search.search.token_searcher import FuzzyTokenSearcher
from fuzzy_search.search.token_searcher import MatchType
from fuzzy_search.search.token_searcher import PartialPhraseMatch
from fuzzy_search.search.token_searcher import get_token_skipgram_matches
from fuzzy_search.search.token_searcher import get_token_skip_match_type
from fuzzy_search.search.token_searcher import get_partial_phrases
from fuzzy_search.tokenization.string import text2skipgrams
from fuzzy_search.tokenization.token import Tokenizer
from fuzzy_search.tokenization.vocabulary import Vocabulary


class TestTokenSearcher(TestCase):

    def test_token_searcher_has_default_tokenizer(self):
        token_searcher = FuzzyTokenSearcher()
        self.assertEqual(True, isinstance(token_searcher.tokenizer, Tokenizer))

    def test_token_searcher_has_token_skipgram_index(self):
        phrase_list = ['test']
        token_searcher = FuzzyTokenSearcher(phrase_list=phrase_list)
        self.assertEqual(True, phrase_list[0] in token_searcher.token_skipgram_index['te'])


def get_token_match_types(text_token, token_searcher):
    text_token_skips = [skipgram for skipgram in text2skipgrams(text_token.normalised_string,
                                                                token_searcher.ngram_size,
                                                                token_searcher.skip_size)]
    text_token_num_skips = len(text_token_skips)
    token_skip_matches = get_token_skipgram_matches(text_token, text_token_skips, token_searcher)
    match_types = []
    for match in token_skip_matches.match_start_offsets:
        print('text_token:', text_token, '\tmatch:', match)
        match_type = get_token_skip_match_type(text_token.normalised_string, text_token_num_skips,
                                               token_skip_matches, match, token_searcher, debug=0)
        match_types.append(match_type)
    return match_types


class TestTokenSearcherMatchType(TestCase):

    def setUp(self) -> None:
        phrase_list = ['test']
        self.text = 'This is a test in which we are testing'
        self.tokenizer = Tokenizer()
        self.doc = self.tokenizer.tokenize_doc(doc_text=self.text)
        self.token_searcher = FuzzyTokenSearcher(phrase_list=phrase_list, tokenizer=self.tokenizer)

    def test_get_match_type_finds_full_match(self):
        text_token = self.doc.tokens[3]
        match_types = get_token_match_types(text_token, self.token_searcher)
        self.assertEqual(True, MatchType.FULL in match_types)

    def test_get_match_type_finds_partial_phrase_token_match(self):
        text_token = self.doc.tokens[3]
        match_types = get_token_match_types(text_token, self.token_searcher)
        self.assertEqual(True, MatchType.FULL in match_types)

    def test_get_match_type_finds_partial_text_token_match(self):
        text_token = self.doc.tokens[8]
        match_types = get_token_match_types(text_token, self.token_searcher)
        self.assertEqual(True, MatchType.PARTIAL_OF_TEXT_TOKEN in match_types)

    def test_token_searcher_finds_multiple_matches(self):
        token_matches = self.token_searcher.find_skipgram_token_matches_in_text(self.doc)
        self.assertEqual(2, len(token_matches))

    def test_token_searcher_finds_partial_phrase_token_match(self):
        self.token_searcher = FuzzyTokenSearcher(phrase_list=['test'], tokenizer=self.tokenizer)
        self.doc = self.tokenizer.tokenize_doc(doc_text='The purpose is testing')
        token_matches = self.token_searcher.find_skipgram_token_matches_in_text(self.doc, debug=0)
        self.assertEqual(MatchType.PARTIAL_OF_TEXT_TOKEN, token_matches[0].match_type)

    def test_get_match_type_finds_multi_text_token_match(self):
        self.token_searcher = FuzzyTokenSearcher(phrase_list=['testing'], tokenizer=self.tokenizer)
        self.doc = self.tokenizer.tokenize_doc(doc_text='We are test ing')
        token_matches = self.token_searcher.find_skipgram_token_matches_in_text(self.doc)
        self.assertEqual((self.doc.tokens[2], self.doc.tokens[3]), token_matches[0].text_tokens)

    def test_get_match_type_finds_same_text_token_match_for_multi_phrase_tokens(self):
        self.token_searcher = FuzzyTokenSearcher(phrase_list=['test', 'case'], tokenizer=self.tokenizer)
        self.doc = self.tokenizer.tokenize_doc(doc_text='This is a testcase')
        token_matches = self.token_searcher.find_skipgram_token_matches_in_text(self.doc)
        for ti, token_match in enumerate(token_matches):
            with self.subTest(ti):
                self.assertEqual(MatchType.PARTIAL_OF_TEXT_TOKEN, token_match.match_type)


class TestPartialPhraseMatch(TestCase):

    def setUp(self) -> None:
        self.tokenizer = Tokenizer()
        self.token_searcher = FuzzyTokenSearcher(phrase_list=['test case'], tokenizer=self.tokenizer)
        self.text = 'This is a test case'
        self.phrase = Phrase(self.text, tokenizer=self.tokenizer)

    def test_making_partial_match_extends_text_tokens(self):
        doc = self.tokenizer.tokenize_doc(doc_text=self.text)
        token_matches = self.token_searcher.find_skipgram_token_matches_in_text(doc)
        partial_match = PartialPhraseMatch(self.phrase, token_matches)
        self.assertEqual(len(partial_match.text_tokens), sum(len(tm.text_tokens) for tm in token_matches))

    def test_making_partial_match_extends_phrase_tokens(self):
        doc = self.tokenizer.tokenize_doc(doc_text='This is a test case')
        token_matches = self.token_searcher.find_skipgram_token_matches_in_text(doc)
        partial_match = PartialPhraseMatch(self.phrase, token_matches)
        self.assertEqual(len(partial_match.phrase_tokens), sum(len(tm.phrase_tokens) for tm in token_matches))

    def test_making_partial_match_extends_text_length(self):
        doc = self.tokenizer.tokenize_doc(doc_text='This is a test case')
        token_matches = self.token_searcher.find_skipgram_token_matches_in_text(doc)
        partial_match = PartialPhraseMatch(self.phrase, token_matches)
        text_length = token_matches[-1].text_end - token_matches[0].text_start
        self.assertEqual(partial_match.text_length, text_length)

    def test_making_partial_match_can_pop(self):
        doc = self.tokenizer.tokenize_doc(doc_text='This is a test case')
        token_matches = self.token_searcher.find_skipgram_token_matches_in_text(doc)
        partial_match = PartialPhraseMatch(self.phrase, token_matches)
        partial_match.pop()
        self.assertEqual(len(partial_match.phrase_tokens), sum(len(tm.phrase_tokens) for tm in token_matches) - 1)

    def test_making_partial_match_picks_best_phrase_token_option(self):
        phrase = Phrase('This is a best test case')
        self.token_searcher = FuzzyTokenSearcher(phrase_list=['best test case'], tokenizer=self.tokenizer)
        doc = self.tokenizer.tokenize_doc(doc_text='This is a best test case')
        token_matches = self.token_searcher.find_skipgram_token_matches_in_text(doc)
        for tm in token_matches:
            print('token_match:', tm)
        # partial_match = PartialPhraseMatch(token_matches, phrase=phrase)
        candidates = get_partial_phrases(token_matches, self.token_searcher)
        self.assertEqual(1, len(candidates))
        for pi, phrase in enumerate(candidates):
            print('test_search_token_searcher - phrase:', phrase)
            with self.subTest(pi):
                for partial in candidates[phrase]:
                    print('\ntest_search_token_searcher - partial:', partial)
                    self.assertEqual(0, len(partial.missing_tokens))
                self.assertEqual(2, len(candidates[phrase]))

    def test_making_partial_match_does_not_duplicate_text_tokens(self):
        phrase = Phrase('This is a best testcase', tokenizer=self.tokenizer)
        doc = self.tokenizer.tokenize_doc(doc_text='This is a best testcase')
        token_matches = self.token_searcher.find_skipgram_token_matches_in_text(doc, debug=0)
        print('phrase:', phrase)
        for tm in token_matches:
            print('\ttoken_match:', tm)
        partial_match = PartialPhraseMatch(phrase, token_matches)
        print('partial_match.text_tokens:', partial_match.text_tokens)
        self.assertEqual(2, len(partial_match.text_tokens))

    def test_making_partial_match_discards_when_missing_phrase_token(self):
        text = 'This is a best test sentence'
        phrase = Phrase(text, tokenizer=self.tokenizer)
        self.token_searcher = FuzzyTokenSearcher(phrase_list=['test case'], tokenizer=self.tokenizer)
        doc = self.tokenizer.tokenize_doc(doc_text=text)
        token_matches = self.token_searcher.find_skipgram_token_matches_in_text(doc, debug=0)
        for tm in token_matches:
            print(tm)
        partial_match = PartialPhraseMatch(phrase, token_matches)
        self.assertEqual(3, len(partial_match.text_tokens))

    def test_making_candidate_phrases_can_detect_term_swap(self):
        phrase = Phrase('best test case', tokenizer=self.tokenizer)
        self.token_searcher = FuzzyTokenSearcher(phrase_list=[phrase], tokenizer=self.tokenizer)
        doc = self.tokenizer.tokenize_doc(doc_text='This is a test best case')
        token_matches = self.token_searcher.find_skipgram_token_matches_in_text(doc)
        for tm in token_matches:
            print(tm)
        candidates = get_partial_phrases(token_matches, self.token_searcher, debug=0)
        for pi, phrase in enumerate(candidates):
            print('test_search_token_searcher - phrase:', phrase)
            with self.subTest(pi):
                for partial in candidates[phrase]:
                    print('\ntest_search_token_searcher - partial:', partial)
                    self.assertEqual(0, len(partial.missing_tokens))
                self.assertEqual(2, len(candidates[phrase]))
        # self.assertEqual(2, len(partial_match.text_tokens))


class TestFindMatches(TestCase):

    def setUp(self) -> None:
        self.tokenizer = Tokenizer()
        self.token_searcher = FuzzyTokenSearcher(phrase_list=['test case'], tokenizer=self.tokenizer)
        self.text = 'This is a test case'
        self.phrase = Phrase(self.text, tokenizer=self.tokenizer)

    def test_find_matches_accepts_text_string(self):
        error = None
        try:
            self.token_searcher.find_matches(self.text)
        except BaseException as err:
            error = err
        self.assertEqual(None, error)

    def test_find_matches_accepts_text_dictionary(self):
        error = None
        try:
            self.token_searcher.find_matches({'text': self.text, 'id': 'some_id'})
        except BaseException as err:
            error = err
        self.assertEqual(None, error)

    def test_find_matches_accepts_text_doc(self):
        error = None
        doc = self.tokenizer.tokenize_doc(self.text, doc_id='some_id')
        try:
            self.token_searcher.find_matches(doc)
        except BaseException as err:
            error = err
        self.assertEqual(None, error)

    def test_making_partial_match_picks_best_phrase_option(self):
        self.token_searcher = FuzzyTokenSearcher(phrase_list=['best case', 'test case'], tokenizer=self.tokenizer)
        doc = self.tokenizer.tokenize_doc(doc_text='This is a best test case')
        phrase_matches = self.token_searcher.find_matches(doc)
        for pi, phrase_match in enumerate(phrase_matches):
            print('test_search_token_searcher - phrase_match:', phrase_match)
            print('\t', phrase_match.levenshtein_similarity)
        self.assertEqual(1, len(phrase_matches))

    def test_find_phrase_matches_finds_best_option_per_text_range(self):
        self.token_searcher = FuzzyTokenSearcher(phrase_list=['best test case'], tokenizer=self.tokenizer)
        doc = self.tokenizer.tokenize_doc(doc_text='this is a best test case to test best case matching')
        phrase_matches = self.token_searcher.find_matches(doc, debug=0)
        for pi, phrase_match in enumerate(phrase_matches):
            print('test_search_token_searcher - phrase_match:', phrase_match)
            print('\t', phrase_match.levenshtein_similarity)
            print('\t', phrase_match.offset, phrase_match.end)
        self.assertEqual(2, len(phrase_matches))

    def test_find_phrase_matches_filters_on_score_threshold(self):
        self.token_searcher = FuzzyTokenSearcher(phrase_list=['best test case'], tokenizer=self.tokenizer)
        doc = self.tokenizer.tokenize_doc(doc_text='this is a best test case to test best case matching')
        self.token_searcher.config['levenshtein_threshold'] = 0.9
        phrase_matches = self.token_searcher.find_matches(doc, debug=0)
        for pi, phrase_match in enumerate(phrase_matches):
            print('test_search_token_searcher - phrase_match:', phrase_match)
            print('\t', phrase_match.levenshtein_similarity)
            print('\t', phrase_match.offset, phrase_match.end)
        for pi, pm in enumerate(phrase_matches):
            with self.subTest(pi):
                self.assertEqual(True, pm.levenshtein_similarity >= self.token_searcher.config['levenshtein_threshold'])

    def test_get_token_skipgram_matches_skips_beyond_max_start(self):
        phrase = {'phrase': 'test', 'max_start_offset': 1}
        self.token_searcher = FuzzyTokenSearcher(phrase_list=[phrase], tokenizer=self.tokenizer)
        doc = self.tokenizer.tokenize_doc(doc_text='a document for a test')
        text_token = doc.tokens[4]
        text_token_skips = [sg for sg in text2skipgrams(text_token.normalised_string)]
        token_skip_matches = get_token_skipgram_matches(text_token, text_token_skips, self.token_searcher)
        self.assertEqual(0, len(token_skip_matches.matches))

    def test_get_token_skipgram_matches_does_not_skip_before_max_start(self):
        phrase = {'phrase': 'test', 'max_start_offset': 25}
        token_searcher = FuzzyTokenSearcher(phrase_list=[phrase], tokenizer=self.tokenizer)
        doc = self.tokenizer.tokenize_doc(doc_text='a document for a test')
        text_token = doc.tokens[4]
        text_token_skips = [sg for sg in text2skipgrams(text_token.normalised_string)]
        token_skip_matches = get_token_skipgram_matches(text_token, text_token_skips, token_searcher)
        self.assertEqual(1, len(token_skip_matches.matches))


class TestTokenSearcherVocabulary(unittest.TestCase):

    def setUp(self) -> None:
        self.vocab = Vocabulary(['best', 'test'])
        self.tokenizer = Tokenizer()

    def test_searcher_can_have_vocab(self):
        token_searcher = FuzzyTokenSearcher(phrase_list=['best test case'], tokenizer=self.tokenizer,
                                            vocabulary=self.vocab)
        self.assertEqual(True, isinstance(token_searcher.vocabulary, Vocabulary))

    def test_searcher_can_match_tokens_via_vocab(self):
        token_searcher = FuzzyTokenSearcher(phrase_list=['best test case'], tokenizer=self.tokenizer,
                                            vocabulary=self.vocab)
        doc = self.tokenizer.tokenize_doc('this is a test best case')
        matches = token_searcher.find_matches(doc)
        self.assertEqual(0, len(matches))


class TestTokenSearcherVariants(unittest.TestCase):

    def setUp(self) -> None:
        self.tokenizer = Tokenizer()
        self.phrase_model = [
            {
                'phrase': 'best',
                'variants': ['rest']
            }
        ]
        self.searcher = FuzzyTokenSearcher(phrase_model=self.phrase_model)

    def test_searcher_can_match_phrase_variants(self):
        # TO DO: check that variants are matched and returned with their main phrase as match phrase
        doc = self.tokenizer.tokenize_doc('this is a time to rest')
        matches = self.searcher.find_matches(doc)
        self.assertEqual(('best', 'rest'), (matches[0].phrase.phrase_string, matches[0].variant.phrase_string))

    """
"""



from __future__ import annotations
import uuid
import string
from collections import defaultdict
from datetime import datetime
from enum import Enum
from typing import Dict, List, Union

import fuzzy_search
import fuzzy_search.tokenization.string as fuzzy_string
from fuzzy_search.match.candidate_match import Candidate
from fuzzy_search.phrase.phrase import Phrase
from fuzzy_search.phrase.phrase_model import PhraseModel
from fuzzy_search.tokenization.token import Token


def filter_matches_by_overlap(filtered_matches: List[PhraseMatch], first_best: bool = False,
                              debug: int = 0) -> List[PhraseMatch]:
    """Filter matches by overlapping match string offsets. When there are multiple phrases matching
    with the same character range in the input text, only pick the matches with the highest
    similarity scores. By default, all matches with the highest similarity score are returned.
    Use 'first_best=True' to return only the first best scoring match.
    """
    if debug > 1:
        print(f"phrase_match.filter_matches_by_overlap - filtered_matches: {len(filtered_matches)}")
    sorted_matches = sorted(filtered_matches, key=lambda x: (x.offset, len(x.string)))
    filtered_matches = []
    if debug > 1:
        print(f"phrase_match.filter_matches_by_overlap - sorted_matches: {len(sorted_matches)}")
    overlapping = defaultdict(list)
    if debug > 1:
        print(f"phrase_match.filter_matches_by_overlap - using first_best: {first_best}")
    for match in sorted_matches:
        overlapping[(match.offset, len(match.string))].append(match)
    for offset_length in overlapping:
        if len(overlapping[offset_length]) == 1:
            filtered_matches.extend(overlapping[offset_length])
        else:
            if first_best is True:
                first_best = max(overlapping[offset_length], key=lambda item: item.levenshtein_similarity)
                filtered_matches.append(first_best)
            else:
                sorted_matches = sorted(overlapping[offset_length], key=lambda item: item.levenshtein_similarity,
                                        reverse=True)
                best_sim = sorted_matches[0].levenshtein_similarity
                if debug > 1:
                    print(f"phrase_match.filter_matches_by_overlap - best similarity score: {best_sim}")
                for best_match in sorted_matches:
                    if best_match.levenshtein_similarity < best_sim:
                        break
                    if debug > 1:
                        print(f"phrase_match.filter_matches_by_overlap - best match: "
                              f"({offset_length})\t{best_match.phrase.phrase_string}")
                    filtered_matches.append(best_match)
    return filtered_matches


def candidates_to_matches(candidates: List[Candidate], text: dict, phrase_model: PhraseModel,
                          ignorecase: bool = False) -> List[PhraseMatch]:
    matches: List[PhraseMatch] = []
    for candidate in candidates:
        if candidate.phrase.phrase_string in phrase_model.is_variant_of:
            match_phrase_string = phrase_model.is_variant_of[candidate.phrase.phrase_string]
            match_phrase = phrase_model.phrase_index[match_phrase_string]
        else:
            match_phrase = candidate.phrase
        # print('candidates_to_matches - ignorecase:', ignorecase)
        match = PhraseMatch(match_phrase, candidate.phrase,
                            candidate.match_string, candidate.match_start_offset, text_id=text["id"],
                            ignorecase=ignorecase,
                            # match_label=match_phrase.label
                            )
        match.add_scores(skipgram_overlap=candidate.get_skip_count_overlap())
        matches.append(match)
    return matches


def validate_match_props(match_phrase: Phrase, match_variant: Phrase,
                         match_string: str, match_offset: int) -> None:
    """Validate match properties.

    :param match_phrase: the phrase that has been matched
    :type match_phrase: Phrase
    :param match_variant: the variant of the phrase that the match is based on
    :type match_variant: Phrase
    :param match_string: the text string that matches the variant phrase
    :type match_string: str
    :param match_offset: the offset of the match string in the text
    :type match_offset: int
    :return: None
    :rtype: None
    """
    if not isinstance(match_phrase, Phrase):
        print(f"match_phrase: {match_phrase}")
        print(f"type: {type(match_phrase)}")
        raise TypeError('match_phrase MUST be of class Phrase')
    if not isinstance(match_variant, Phrase):
        raise TypeError('match_variant MUST be of class Phrase')
    if not isinstance(match_string, str):
        raise TypeError('match string MUST be a string')
    if len(match_string) == 0:
        print(f"match_phrase: {match_phrase}")
        raise ValueError('match string cannot be empty string')
    if not isinstance(match_offset, int):
        raise TypeError('match_offset must be an integer')
    if match_offset < 0:
        raise ValueError('offset cannot be negative')


def adjust_match_start_offset(text: Dict[str, any], match_string: str,
                              match_offset: int) -> Union[int, None]:
    """Adjust the start offset if it is not at a word boundary.

    :param text: the text object that contains the candidate match string
    :type text: Dict[str, any]
    :param match_string: the candidate match string
    :type match_string: str
    :param match_offset: the text offset of the candidate match string
    :type match_offset: int
    :return: the adjusted offset or None if the required adjustment is too big
    :rtype: Union[int, None]
    """
    # adjust the start
    # check if there match initial is a non-word character
    non_word_prefix = fuzzy_string.get_non_word_prefix(match_string)
    # print('non_word_prefix:', non_word_prefix)
    if non_word_prefix == "":
        # match does not start with a non-word prefix, so check if it needs to be moved to the left
        if match_offset == 0:
            # match is at the start of text and starts with word characters
            return match_offset
        # if character before match is first of text and not a word boundary, move left
        if match_offset == 1 and text['text'][0] not in fuzzy_string.non_word_affixes_1:
            return match_offset - 1
        # if character before match is a word boundary, match offset is good
        if text["text"][match_offset-1:match_offset] in fuzzy_string.non_word_affixes_1:
            return match_offset
        # if penultimate character before match is a word boundary, move offset by -1
        elif match_offset > 1 and text["text"][match_offset-2:match_offset-1] in fuzzy_string.non_word_affixes_1:
            # move match_offset back by 1 to start at word boundary
            return match_offset-1
        # if penultimate character before match is start of text, move offset by -2
        elif match_offset == 2 and text['text'][0] not in fuzzy_string.non_word_affixes_1:
            return match_offset - 2
        # if two characters before match is a word boundary, move offset by -2
        elif match_offset > 2 and text["text"][match_offset-3:match_offset-2] in fuzzy_string.non_word_affixes_1:
            # move match_offset back by 1 to start at word boundary
            return match_offset-2
        # if the three characters preceding match are word characters, the start is wrong
        else:
            return None
    else:
        # match starts with a non-word-prefix, so move offset to after the prefix
        return match_offset + len(non_word_prefix)


def adjust_match_end_offset(phrase_string: str, candidate_string: str,
                            text: Dict[str, any], end_offset: int, punctuation: str,
                            debug: int = 0) -> Union[int, None]:
    """Adjust the end offset if it is not at a word boundary.

    :param phrase_string: the phrase string
    :type phrase_string: str
    :param candidate_string: the candidate match string
    :type candidate_string: str
    :param text: the text object that contains the candidate match string
    :type text: Dict[str, any]
    :param end_offset: the text offset of the candidate match string
    :type end_offset: int
    :param punctuation: the set of characters to treat as punctuation
    :type punctuation: str
    :param debug: level to show debug information
    :type debug: int
    :return: the adjusted offset or None if the required adjustment is too big
    :rtype: Union[int, None]
    """
    # ugly hack: if phrase string ends with punctuation, use only whitespace as word end boundary
    if debug > 2:
        print('\tadjust_match_end_offset - start')
    whitespace_only = True if phrase_string[-1] in punctuation else False
    if debug > 2:
        print('\tadjust_match_end_offset - whitespace_only:', whitespace_only)
    phrase_end = map_string(phrase_string[-3:], punctuation)
    if debug > 2:
        print('\tadjust_match_end_offset - prhase_end:', phrase_end)
    match_end = map_string(candidate_string[-3:], punctuation, whitespace_only=whitespace_only)
    if debug > 2:
        print('\tadjust_match_end_offset - match_end:', match_end)
    text_suffix = map_string(text["text"][end_offset:end_offset+3], punctuation,
                             whitespace_only=whitespace_only, debug=debug)
    if debug > 2:
        print('\tadjust_match_end_offset - text_suffix:', text_suffix)
        print(f"\tadjust_match_end_offset - match_end: {candidate_string[-3:]: <4}\ttext_suffix: {text['text'][end_offset:end_offset+3]: >4}")
        print(f"\tadjust_match_end_offset - mapped suffixes - match_end: #{match_end}#\ttext_suffix: #{text_suffix}#")
    return calculate_end_shift(phrase_end, match_end, text_suffix, end_offset)


def adjust_match_offsets(phrase_string: str, candidate_string: str,
                         text: Dict[str, any], candidate_start_offset: int,
                         candidate_end_offset: int,
                         punctuation: str = string.punctuation,
                         debug: int = 0) -> Union[Dict[str, Union[str, int]], None]:
    """Adjust the end offset if it is not at a word boundary.

    :param phrase_string: the phrase string
    :type phrase_string: str
    :param candidate_string: the candidate match string
    :type candidate_string: str
    :param text: the text object that contains the candidate match string
    :type text: Dict[str, any]
    :param candidate_start_offset: the text offset of the start of the candidate match string
    :type candidate_start_offset: int
    :param candidate_end_offset: the text offset of the end of the candidate match string
    :type candidate_end_offset: int
    :param punctuation: the set of characters to treat as punctuation (defaults to string.punctuation)
    :type punctuation: str
    :param debug: level to show debug information
    :type debug: int
    :return: the adjusted offset or None if the required adjustment is too big
    :rtype: Union[int, None]
    """
    if debug > 2:
        print("\tadjust_match_offset - phrase string:", phrase_string)
        print("\tadjust_match_offset - adjusting candidate string:", candidate_string)
    if punctuation is None:
        punctuation = string.punctuation
    if debug > 2:
        print("\tadjust_match_offset - candidate_start_offset:", candidate_start_offset)
    match_start_offset = adjust_match_start_offset(text, candidate_string, candidate_start_offset)
    if debug > 2:
        print("\tadjust_match_offset - match_start_offset:", match_start_offset)
    if match_start_offset is None:
        return None
    match_end_offset = adjust_match_end_offset(phrase_string, candidate_string,
                                               text, candidate_end_offset, punctuation, debug=debug)
    if debug > 2:
        print("\tadjust_match_offset - match_end_offset:", match_end_offset)
    if match_end_offset is None:
        return None
    elif match_end_offset <= match_start_offset:
        return None
    return {
        "match_string": text["text"][match_start_offset:match_end_offset],
        "match_start_offset": match_start_offset,
        "match_end_offset": match_end_offset
    }


def map_string(affix_string: str, punctuation: str,
               whitespace_only: bool = False, debug: int = 0) -> str:
    """Turn affix string into type char representation. Types are 'w' for non-whitespace char,
    and 's' for whitespace char.

    :param affix_string: a string
    :type: str
    :param punctuation: the set of characters to treat as punctuation
    :type punctuation: str
    :param whitespace_only: whether to treat only whitespace as word boundary or also include (some) punctuation
    :type whitespace_only: bool
    :param debug: level to show debug information
    :type debug: int
    :return: the type char representation
    :rtype: str
    """
    if whitespace_only:
        return ''.join(['s' if char in ' \t\n\r' else 'w' for char in affix_string])
    else:
        return ''.join(['s' if char in ' \t\n\r' or char in punctuation else 'w' for char in affix_string])


def calculate_end_shift(phrase_end: str, match_end: str, text_suffix: str, end_offset: int):
    if phrase_end == match_end:
        if text_suffix == "" or text_suffix.startswith("s"):
            return end_offset
    if phrase_end.endswith("s") and match_end.endswith("s"):
        return end_offset
    if match_end == "wss":
        return end_offset - 2
    if phrase_end == "www":
        if match_end == "www":
            if text_suffix == "w" or text_suffix.startswith("ws"):
                return end_offset + 1
            elif text_suffix == "ww" or text_suffix.startswith("wws"):
                return end_offset + 2
            elif text_suffix.startswith("www"):
                return None
        if match_end == "wws":
            return end_offset - 1
        if match_end == "wsw":
            if text_suffix == "" or text_suffix.startswith("s"):
                # we assume the whitespace in the match is a misrecognised word character
                return end_offset
            if text_suffix.startswith("w"):
                # we assume the whitespace in the match is correct
                return end_offset - 2
        if match_end == "sww":
            if text_suffix == "" or text_suffix.startswith("s"):
                # we assume the whitespace in the match is a misrecognised word character
                return end_offset
            elif text_suffix.startswith("w"):
                # we assume the whitespace in the match is correct
                return None
        if match_end == "sws":
            # we assume the first whitespace in the match is a misrecognised word character
            return end_offset - 1
        if match_end == "ssw":
            return None
        else:
            return None
    if phrase_end == "wws":
        if match_end == "www":
            if text_suffix == "":
                return end_offset
            elif text_suffix.startswith("s"):
                return end_offset + 1
            elif text_suffix == "w":
                return None
            elif text_suffix.startswith("ws"):
                return end_offset + 2
            elif text_suffix.startswith("ww"):
                return None
            else:
                return None
        elif match_end.startswith("ws"):
            return end_offset - 1
        elif match_end.startswith("s"):
            return end_offset - 2
        else:
            return None
    if phrase_end == "sww":
        if match_end == "sww":
            if text_suffix == "w" or text_suffix.startswith("ws"):
                return end_offset + 1
            else:
                return None
        elif match_end == "sws":
            return end_offset - 1
        elif match_end == "www":
            if text_suffix == "" or text_suffix.startswith("s"):
                return end_offset
            else:
                return None
        elif match_end == "wsw":
            if text_suffix == "" or text_suffix.startswith("s"):
                return end_offset
            if text_suffix == "ws" or text_suffix.startswith("ws"):
                return end_offset + 1
            if text_suffix == "ww" or text_suffix.startswith("wws"):
                return end_offset + 2
            else:
                return None
        elif match_end == "ssw":
            if text_suffix == "" or text_suffix.startswith("s"):
                return end_offset
            elif text_suffix == "w" or text_suffix.startswith("ws"):
                return end_offset + 1
            elif text_suffix == "ww" or text_suffix.startswith("wws"):
                return end_offset + 1
            else:
                return None
        else:
            return None
    if phrase_end == "sws":
        if match_end == "www":
            if text_suffix == "sw" or text_suffix == "sws":
                return end_offset + 2
            else:
                return None
        elif match_end == "sww":
            return end_offset - 2
        elif match_end == "wsw":
            if text_suffix == "":
                return end_offset
            if text_suffix.startswith("s"):
                return end_offset + 1
            else:
                return end_offset - 1
        else:
            return None
    if phrase_end == "wsw":
        if match_end == "wsw":
            if text_suffix == "w" or text_suffix.startswith("ws"):
                return end_offset + 1
            else:
                return None
        if match_end == "www":
            if text_suffix == "" or text_suffix == "s":
                return end_offset
            elif text_suffix.startswith("w"):
                return None
            elif text_suffix == "sw" or text_suffix == "sws":
                return end_offset + 2
            else:
                return None
        if match_end == "sww":
            if text_suffix == "":
                return end_offset + 1
            elif text_suffix.startswith("s"):
                return end_offset
            else:
                return None
        if match_end == "ssw":
            if text_suffix == "" or text_suffix.startswith("s"):
                return end_offset
            else:
                return None
        else:
            return None
    if len(phrase_end) < 3:
        if phrase_end == match_end:
            return end_offset
        else:
            return None
    else:
        details = f"phrase_end {phrase_end}, match_end {match_end}, text_suffix {text_suffix}"
        raise ValueError(f"combination not captured: {details}")


###############
# Match class #
###############

class PhraseMatch:
    """

    Attributes
    """

    def __init__(self, match_phrase: Phrase, match_variant: Phrase, match_string: str,
                 match_offset: int, ignorecase: bool = False, text_id: Union[None, str] = None,
                 match_scores: dict = None, match_label: Union[str, List[str]] = None,
                 match_id: str = None, levenshtein_similarity: float = None):
        """

        :param match_phrase: a phrase object for which a matching string is found in the text
        :param match_variant: a phrase object for the variant that matches the string in the text
        :param match_string: the matching string found in the text
        :param match_offset: the offset of the matching string in the text
        :param ignorecase: boolean flag whether to ignore case differences
        :param text_id: the identifier of the text in which the match is found
        :param match_scores: the similarity scores of the match
        :param match_label: one or more labels to attach to the match
        :param match_id: an optional identifier to use for the match
        """
        # print("Match class match_phrase:", match_phrase)
        validate_match_props(match_phrase, match_variant, match_string, match_offset)
        self.id = match_id if match_id else str(uuid.uuid4())
        self.phrase = match_phrase
        self.label = match_phrase.label
        if match_label:
            self.label = match_label
        # if self.label is None:
        #     print(f'PhraseMatch - self.label is None - match_phrase: {match_phrase}')
        #     print(f'PhraseMatch - self.label is None - match_phrase.label: {match_phrase.label}')
        self.metadata = {}
        self.variant = match_variant
        self.string = match_string
        self.ignorecase = ignorecase
        self.offset = match_offset
        self.end = self.offset + len(self.string)
        self.text_id = text_id
        self.character_overlap: Union[None, float] = None
        self.ngram_overlap: Union[None, float] = None
        self.skipgram_overlap: Union[None, float] = None
        self.levenshtein_similarity: Union[None, float] = levenshtein_similarity
        if match_scores:
            self.character_overlap = match_scores['char_match']
            self.ngram_overlap = match_scores['ngram_match']
            self.levenshtein_similarity = match_scores['levenshtein_similarity']
        self.created = datetime.now()

    def __repr__(self):
        return f'PhraseMatch(' + \
            f'phrase: "{self.phrase.phrase_string}", variant: "{self.variant.phrase_string}", ' + \
            f'string: "{self.string}", offset: {self.offset}, ignorecase: {self.ignorecase}, ' + \
            f'levenshtein_similarity: {self.levenshtein_similarity})'

    @property
    def label_list(self) -> List[str]:
        if isinstance(self.label, str):
            return [self.label]
        elif isinstance(self.label, list):
            return self.label
        else:
            return []

    def has_label(self, label: str):
        if isinstance(self.label, str):
            return label == self.label
        elif isinstance(self.label, list):
            return label in self.label
        else:
            return label in self.label

    def json(self) -> dict:
        data = {
            "type": "PhraseMatch",
            "phrase": self.phrase.phrase_string,
            "variant": self.variant.phrase_string,
            "string": self.string,
            "offset": self.offset,
            "label": self.label,
            "ignorecase": self.ignorecase,
            "text_id": self.text_id,
            "match_scores": {
                "char_match": self.character_overlap,
                "ngram_match": self.ngram_overlap,
                "levenshtein_similarity": self.levenshtein_similarity
            }
        }
        if "label" in self.phrase.metadata:
            data["label"] = self.phrase.metadata["label"]
        return data

    @staticmethod
    def from_json(match_json):
        match_phrase = Phrase(phrase=match_json['phrase'])
        match_variant = Phrase(phrase=match_json['variant'])
        return PhraseMatch(match_phrase=match_phrase, match_variant=match_variant,
                           match_string=match_json['string'], match_offset=match_json['offset'],
                           text_id=match_json['text_id'], match_scores=match_json['match_scores'],
                           match_label=match_json['label'], ignorecase=match_json['ignorecase'])

    def add_scores(self, skipgram_overlap: Union[None, float] = None) -> None:
        """Compute overlap and similarity scores between the match variant and the match string
        and add these to the match object.

        :param skipgram_overlap: the overlap in skipgrams between match string and match variant
        :type skipgram_overlap: Union[float, None]
        :return: None
        :rtype: None
        """
        # print('PhraseMatch - ignorecase:', self.ignorecase)
        match_string = self.string.lower() if self.ignorecase else self.string
        phrase_string = self.variant.phrase_string.lower() if self.ignorecase else self.variant.phrase_string
        # print('match_string:', match_string)
        # print('variant.phrase_string:', self.variant.phrase_string)
        self.character_overlap = fuzzy_string.score_char_overlap_ratio(phrase_string, match_string)
        self.ngram_overlap = fuzzy_string.score_ngram_overlap_ratio(phrase_string, match_string,
                                                                    self.variant.ngram_size)
        self.levenshtein_similarity = fuzzy_string.score_levenshtein_similarity_ratio(phrase_string,
                                                                                      match_string)
        if skipgram_overlap is not None:
            self.skipgram_overlap = skipgram_overlap

    def score_character_overlap(self):
        """Return the character overlap between the variant phrase_string and the match_string

        :return: the character overlap as proportion of the variant phrase string
        :rtype: float
        """
        match_string = self.string.lower() if self.ignorecase else self.string
        phrase_string = self.variant.phrase_string.lower() if self.ignorecase else self.variant.phrase_string
        # print('match_string:', match_string)
        # print('variant.phrase_string:', self.variant.phrase_string)
        self.character_overlap = fuzzy_string.score_char_overlap_ratio(phrase_string, match_string)
        return self.character_overlap

    def score_ngram_overlap(self) -> float:
        """Return the ngram overlap between the variant phrase_string and the match_string

        :return: the ngram overlap as proportion of the variant phrase string
        :rtype: float
        """
        match_string = self.string.lower() if self.ignorecase else self.string
        phrase_string = self.variant.phrase_string.lower() if self.ignorecase else self.variant.phrase_string
        # print('match_string:', match_string)
        # print('variant.phrase_string:', self.variant.phrase_string)
        self.ngram_overlap = fuzzy_string.score_ngram_overlap_ratio(phrase_string,
                                                                    match_string, self.phrase.ngram_size)
        return self.ngram_overlap

    def score_levenshtein_similarity(self):
        """Return the levenshtein similarity between the variant phrase_string and the match_string

        :return: the levenshtein similarity as proportion of the variant phrase string
        :rtype: float
        """
        match_string = self.string.lower() if self.ignorecase else self.string
        phrase_string = self.variant.phrase_string.lower() if self.ignorecase else self.variant.phrase_string
        # print('match_string:', match_string)
        # print('variant.phrase_string:', self.variant.phrase_string)
        self.levenshtein_similarity = fuzzy_string.score_levenshtein_similarity_ratio(phrase_string,
                                                                                      match_string)
        return self.levenshtein_similarity

    def overlaps(self, other: PhraseMatch) -> bool:
        """Check if the match string of this match object overlaps with the match string of another match object.

        :param other: another match object
        :type other: PhraseMatch
        :return: a boolean indicating whether the match_strings of the two objects overlap in the source text
        :rtype: bool"""
        if self.text_id is not None and self.text_id != other.text_id:
            return False
        if self.offset <= other.offset < self.end:
            return True
        elif other.offset <= self.offset < other.end:
            return True
        else:
            return False

    def as_web_anno(self) -> Dict[str, any]:
        """Turn match object into a W3C Web Annotation representation"""
        if not self.text_id:
            raise ValueError('Cannot make target: match object has no text_id')
        body_match = [
            {
                'type': 'TextualBody',
                'purpose': 'tagging',
                'format': 'text',
                'value': self.phrase.phrase_string
            },
            {
                'type': 'TextualBody',
                'purpose': 'highlighting',
                'format': 'text',
                'value': self.string
            }
        ]
        if self.variant.phrase_string != self.string:
            correction = {
                'type': 'TextualBody',
                'purpose': 'correcting',
                'format': 'text',
                'value': self.variant.phrase_string
            }
            body_match.append(correction)
        if self.label:
            classification = {
                'type': 'TextualBody',
                'purpose': 'classifying',
                'format': 'text',
                'value': self.label
            }
            body_match.append(classification)
        return {
            "@context": "http://www.w3.org/ns/anno.jsonld",
            "id": self.id,
            "type": "Annotation",
            "motivation": "classifying",
            "created": self.created.isoformat(),
            "generator": {
                "id": "https://github.com/marijnkoolen/fuzzy-search",
                "type": "Software",
                "name": f"fuzzy-search v{fuzzy_search.__version__}"
            },
            "target": {
                "source": self.text_id,
                "selector": {
                    "type": "TextPositionSelector",
                    "start": self.offset,
                    "end": self.end
                }
            },
            "body": body_match
        }


class PhraseMatchInContext(PhraseMatch):

    def __init__(self, match: PhraseMatch, text: Union[str, dict] = None, context: str = None,
                 context_start: int = None, context_end: int = None,
                 prefix_size: int = 20, suffix_size: int = 20):
        super().__init__(match_phrase=match.phrase, match_variant=match.variant, match_string=match.string,
                         match_offset=match.offset, text_id=match.text_id)
        """MatchInContext extends a Match object with surrounding context from the text document that the match
        phrase was taken from. Alternatively, the context can be submitted.

        :param text: the text (string or dictionary with 'text' and 'id' properties) that the match phrase was taken from
        :type text: Union[str, dict]
        :param context: the context string around the match phrase 
        :type context: Union[str, dict]
        :param match: the match phrase object
        :type match: Match
        :param context_start: the start offset of the context in the original text
        :type context_start: int
        :param context_end: the end offset of the context in the original text
        :type context_end: int
        :param prefix_size: the size of the prefix window
        :type prefix_size: int
        :param suffix_size: the size of the suffix window
        :type suffix_size: int 
        """
        self.character_overlap = match.character_overlap
        self.ngram_overlap = match.ngram_overlap
        self.levenshtein_similarity = match.levenshtein_similarity
        self.prefix_size = prefix_size
        self.suffix_size = suffix_size
        if text:
            if isinstance(text, str):
                text = {"text": text, "id": match.text_id}
            self.context_start = match.offset - prefix_size if match.offset >= prefix_size else 0
            self.context_end = match.end + suffix_size if len(text["text"]) > match.end + suffix_size else len(text["text"])
            self.context = text["text"][self.context_start:self.context_end]
        elif context:
            self.context = context
            self.context_start = context_start
            self.context_end = context_end
        self.prefix = text["text"][self.context_start:match.offset]
        self.suffix = text["text"][match.end:self.context_end]

    def __repr__(self):
        return f'PhraseMatchInContext(' + \
               f'phrase: "{self.phrase.phrase_string}", variant: "{self.variant.phrase_string}",' + \
               f'string: "{self.string}", offset: {self.offset}), context: "{self.context}"'

    def json(self):
        json_data = super().json()
        json_data["context_start"] = self.context_start
        json_data["context_end"] = self.context_end
        json_data["context"] = self.context
        json_data['prefix_size'] = self.prefix_size
        json_data['suffix_size'] = self.suffix_size
        json_data['prefix'] = self.prefix
        json_data['suffix'] = self.suffix
        return json_data

    def as_web_anno(self) -> Dict[str, any]:
        match_anno = super().as_web_anno()
        position_selector = match_anno['target']['selector']
        quote_selector = {
            'type': 'TextQuoteSelector',
            'prefix': self.prefix,
            'exact': self.string,
            'suffix': self.suffix
        }
        match_anno['target']['selector'] = [position_selector, quote_selector]
        return match_anno


def phrase_match_from_json(match_json: dict) -> PhraseMatch:
    match_phrase = Phrase(match_json['phrase'])
    match_variant = Phrase(match_json['variant'])
    phrase_match = PhraseMatch(match_phrase, match_variant, match_json['string'],
                               match_offset=match_json['offset'],
                               match_scores=match_json['match_scores'],
                               match_label=match_json['label'])
    if 'context' in match_json:
        phrase_match = PhraseMatchInContext(phrase_match, context=match_json['context'],
                                            prefix_size=match_json['prefix_size'],
                                            suffix_size=match_json['suffix_size'],
                                            context_start=match_json['context_start'],
                                            context_end=match_json['context_end'])
    return phrase_match


class MatchType(Enum):
    NONE = 0
    PARTIAL_OF_PHRASE_TOKEN = 0.5
    FULL = 1
    PARTIAL_OF_TEXT_TOKEN = 1.5


class TokenMatch:

    def __init__(self, text_tokens: Union[Token, List[Token]],
                 phrase_tokens: Union[str, List[str]],
                 match_type: MatchType):
        if isinstance(text_tokens, Token):
            text_tokens = (text_tokens, )
        elif isinstance(text_tokens, list):
            text_tokens = tuple(text_tokens)
        if isinstance(phrase_tokens, str):
            phrase_tokens = (phrase_tokens, )
        elif isinstance(phrase_tokens, list):
            phrase_tokens = tuple(phrase_tokens)
        self.text_tokens = text_tokens
        self.phrase_tokens = phrase_tokens
        self.match_type = match_type
        self.first = text_tokens[0]
        self.last = text_tokens[-1]
        self.text_start = self.first.char_index
        self.text_end = self.last.char_index + len(self.last)
        self.text_length = self.text_end - self.text_start

    def __repr__(self):
        return f"{self.__class__.__name__}(match_type={self.match_type}, " \
               f"text_tokens={self.text_tokens}, phrase_tokens={self.phrase_tokens})"


class PartialPhraseMatch:

    def __init__(self, phrase: Phrase, token_matches: List[TokenMatch] = None, max_char_gap: int = 20,
                 max_token_gap: int = 1):
        # create a new list instead of pointing to original list
        self.token_matches = []
        self.phrase = phrase
        self.text_tokens = []
        self.phrase_tokens = []
        self.text_phrase_map = defaultdict(list)
        self.missing_tokens = [token.n for token in phrase.tokens]
        self.redundant_tokens = []
        self.max_char_gap = max_char_gap
        self.max_token_gap = max_token_gap
        self.text_start = -1
        self.text_end = -1
        self.text_length = 0
        self.match_string = None
        self.first_text_token = None
        self.last_text_token = None
        self.first_phrase_token = None
        self.last_phrase_token = None
        self.levenshtein_score = None
        if token_matches is not None:
            self.add_tokens(token_matches)

    def __repr__(self):
        return f"{self.__class__.__name__}(phrase={self.phrase}, token_matches={self.token_matches}, " \
               f"text_tokens={self.text_tokens}, phrase_tokens={self.phrase_tokens}," \
               f", missing_tokens={self.missing_tokens})"

    def _update(self):
        text_tokens = []
        prev_match = None
        for match in self.token_matches:
            if prev_match is None:
                text_tokens.extend(match.text_tokens)
            elif match.text_start == prev_match.text_start:
                continue
            elif match.text_start >= prev_match.text_end:
                text_tokens.extend(match.text_tokens)
            else:
                print('TO DO: figure out how to filter text tokens in partially overlapping token matches')
            # print('_update - match.text_start', match.text_start)
            prev_match = match
        # print('text_tokens:', text_tokens)
        self.text_tokens = tuple(text_tokens)
        # self.text_tokens = tuple([token for match in self.token_matches for token in match.text_tokens])
        self.phrase_tokens = tuple([token for match in self.token_matches for token in match.phrase_tokens])
        self.first_text_token = self.text_tokens[0]
        self.last_text_token = self.text_tokens[-1]
        self.text_start = self.first_text_token.char_index
        self.text_end = self.last_text_token.char_index + len(self.last_text_token)
        self.text_length = self.text_end - self.text_start

    def pop(self):
        self.token_matches.pop(0)
        self._update()

    def _check_gap(self, token_match: TokenMatch):
        token_gap = token_match.text_tokens[0].index - self.text_tokens[-1].index
        char_gap = token_match.text_tokens[0].char_index - self.text_end
        if token_gap > self.max_token_gap or char_gap > self.max_char_gap:
            self.__init__(phrase=self.phrase)

    def push(self, token_match: TokenMatch):
        self.token_matches.append(token_match)
        if len(self.text_tokens) > 0:
            self._check_gap(token_match)
        for text_token in token_match.text_tokens:
            self.text_tokens.append(text_token)
            self.text_phrase_map[text_token].extend(list(token_match.phrase_tokens))
        for phrase_token in token_match.phrase_tokens:
            self.phrase_tokens.append(phrase_token)
            if phrase_token in self.missing_tokens:
                self.missing_tokens.remove(phrase_token)
            else:
                self.redundant_tokens.append(phrase_token)

    def add_tokens(self, token_matches: Union[List[TokenMatch], TokenMatch]):
        if isinstance(token_matches, TokenMatch):
            token_matches = [token_matches]
        for token_match in token_matches:
            for phrase_token in token_match.phrase_tokens:
                if phrase_token in self.missing_tokens:
                    self.missing_tokens.remove(phrase_token)
        self.token_matches.extend(token_matches)
        self._update()


def copy_partial_match(partial_match: PartialPhraseMatch):
    new_pm = PartialPhraseMatch(phrase=partial_match.phrase, token_matches=None,
                                max_char_gap=partial_match.max_char_gap,
                                max_token_gap=partial_match.max_token_gap)
    new_pm.token_matches = [tm for tm in partial_match.token_matches]
    new_pm.missing_tokens = [token for token in partial_match.missing_tokens]
    new_pm.text_tokens = [token for token in partial_match.text_tokens]
    new_pm.phrase_tokens = [token for token in partial_match.phrase_tokens]
    new_pm.redundant_tokens = [token for token in partial_match.redundant_tokens]
    new_pm.first_text_token = partial_match.first_text_token
    new_pm.last_text_token = partial_match.last_text_token
    new_pm.text_start = partial_match.text_start
    new_pm.text_end = partial_match.text_end
    new_pm.text_length = partial_match.text_length
    return new_pm



import copy
import string
from collections import defaultdict
from typing import Dict, List, Set, Union

import fuzzy_search
from fuzzy_search.match.skip_match import SkipMatches
from fuzzy_search.match.phrase_match import PhraseMatch
from fuzzy_search.phrase.phrase import Phrase
from fuzzy_search.phrase.phrase_model import PhraseModel
from fuzzy_search.search.config import default_config
from fuzzy_search.tokenization.string import text2skipgrams
from fuzzy_search.tokenization.token import Tokenizer


class FuzzySearcher(object):

    def __init__(self, phrase_list: List[any] = None, phrase_model: Union[Dict[str, any], PhraseModel] = None,
                 config: Union[None, Dict[str, Union[str, int, float]]] = None, tokenizer: Tokenizer = None):
        """This class represents the basic fuzzy searcher. You can pass a list of phrases or a phrase model and
        configuration dictionary that overrides the default configuration values. The default config dictionary
        is available via `fuzzy_search.default_config`.

        To set e.g. the character ngram_size to 3 and the skip_size to 1 use the following dictionary:

        config = {
            'ngram_size': 3,
            'skip_size': 1
        }

        :param phrase_list: a list of phrases (a list of strings or more complex dictionaries with phrases and variants)
        :type phrase_list: list
        :param phrase_model: a phrase model
        :type phrase_model: PhraseModel
        :param config: a configuration dictionary to override default configuration properties.
        Only the properties in the config dictionaries of updated.
        :type config: dict
        :param tokenizer: a tokenizer instance (default tokenizer splits on whitespace)
        :type tokenizer: Tokenizer
        """
        self.__version__ = fuzzy_search.__version__
        # default configuration
        self.char_match_threshold = 0.5
        self.ngram_threshold = 0.5
        self.skipgram_threshold = 0.2
        self.levenshtein_threshold = 0.5
        self.max_length_variance = 1
        self.allow_overlapping_matches = False
        self.skip_exact_matching = False
        self.use_word_boundaries = True
        self.ignorecase = False
        self.known_candidates = defaultdict(dict)
        self.distractor_terms = defaultdict(list)
        self.ngram_size = 2
        self.skipgram_index = defaultdict(set)
        self.early_skipgram_index = defaultdict(set)
        self.late_skipgram_index = defaultdict(set)
        self.skip_size = 2
        self.variant_map = defaultdict(dict)
        self.has_variant = defaultdict(dict)
        self.variant_skipgram_index = defaultdict(set)
        self.variant_early_skipgram_index = defaultdict(set)
        self.variant_late_skipgram_index = defaultdict(set)
        self.include_variants = False
        self.filter_distractors = False
        self.distractor_map = defaultdict(dict)
        self.has_distractor = defaultdict(dict)
        self.distractor_skipgram_index = defaultdict(set)
        self.distractor_early_skipgram_index = defaultdict(set)
        self.distractor_late_skipgram_index = defaultdict(set)
        self.phrases: Set[Phrase] = set()
        self.variants: Set[Phrase] = set()
        self.distractors: Set[Phrase] = set()
        self.phrase_model: Union[None, PhraseModel] = None
        self.debug = False
        self.punctuation = string.punctuation
        # non-default configuration
        self.config = copy.deepcopy(default_config)
        self.tokenizer = tokenizer if tokenizer is not None else Tokenizer()
        if config:
            for key in config:
                self.config[key] = config[key]
            self.configure(config)
        if phrase_list is not None:
            phrase_model = PhraseModel(phrases=phrase_list, config=config, tokenizer=self.tokenizer)
            self.index_phrase_model(phrase_model)
        if phrase_model is not None:
            if isinstance(phrase_model, dict) or isinstance(phrase_model, list):
                phrase_model = PhraseModel(model=phrase_model, config=config, tokenizer=self.tokenizer)
            elif isinstance(phrase_model, PhraseModel) is False:
                raise TypeError('invalid phrase_model type, should PhraseModel or a list of dictionaries')
            self.index_phrase_model(phrase_model)

    def configure(self, config: Dict[str, any]) -> None:
        """Configure the fuzzy searcher with a given config object.

        :param config: a config dictionary
        :type config: Dict[str, Union[str, int, float]]
        """
        if "char_match_threshold" in config:
            self.char_match_threshold = config["char_match_threshold"]
        if "ngram_threshold" in config:
            self.ngram_threshold = config["ngram_threshold"]
        if "skipgram_threshold" in config:
            self.skipgram_threshold = config["skipgram_threshold"]
        if "levenshtein_threshold" in config:
            self.levenshtein_threshold = config["levenshtein_threshold"]
        if "max_length_variance" in config:
            self.max_length_variance = config["max_length_variance"]
        if "use_word_boundaries" in config:
            self.use_word_boundaries = config["use_word_boundaries"]
        if "ignorecase" in config:
            self.ignorecase = config["ignorecase"]
        if "ngram_size" in config:
            self.ngram_size = config["ngram_size"]
        if "skip_size" in config:
            self.skip_size = config["skip_size"]
        if "include_variants" in config:
            self.include_variants = config["include_variants"]
        if "filter_distractors" in config:
            self.filter_distractors = config["filter_distractors"]
        if "skip_exact_matching" in config:
            self.skip_exact_matching = config["skip_exact_matching"]
        if "allow_overlapping_matches" in config:
            self.allow_overlapping_matches = config["allow_overlapping_matches"]
        if "punctuation" in config:
            self.punctuation = config["punctuation"]
        if "debug" in config:
            self.debug = config["debug"]

    def _get_debug_level(self, debug: int = 0):
        if debug > self.debug:
            return debug
        else:
            return self.debug

    def index_phrase_model(self, phrase_model: Union[List[Dict[str, Union[str, int, float, list]]], PhraseModel],
                           debug: int = 0):
        """Add a phrase model to search for phrases in texts.

        :param phrase_model: a phrase model, either as dictionary or as PhraseModel object
        :type phrase_model: Union[List[Dict[str, Union[str, int, float, list]]], PhraseModel]
        :param debug: level to show debug information
        :type debug: int
        """
        debug = self._get_debug_level(debug)
        if isinstance(phrase_model, list):
            phrase_model = PhraseModel(model=phrase_model, config=self.config, tokenizer=self.tokenizer)
        self.phrase_model = phrase_model
        if debug > 3:
            print(f'{self.__class__.__name__}.index_phrase_model - calling index_phrases()')
        self.index_phrases(list(phrase_model.phrase_index.values()))
        if debug > 3:
            print(f'{self.__class__.__name__}.index_phrase_model - calling index_variants()')
        self.index_variants(list(phrase_model.variant_index.values()))
        if debug > 3:
            print(f'{self.__class__.__name__}.index_phrase_model - calling index_distractors()')
        self.index_distractors(list(phrase_model.distractor_index.values()))

    def index_phrases(self, phrases: List[Union[str, Phrase]]) -> None:
        """Add a list of phrases to search for in texts.

        :param phrases: a list of phrases, either as string or as Phrase objects
        :type phrases: List[Union[str, Phrase]]
        """
        for phrase in phrases:
            if isinstance(phrase, str):
                phrase = Phrase(phrase, ngram_size=self.ngram_size, skip_size=self.skip_size)
            if phrase.ngram_size != self.ngram_size:
                searcher_size = f"{self.__class__.__name__} ({self.ngram_size}"
                raise ValueError(f"phrase has different ngram_size ({phrase.ngram_size}) than {searcher_size}")
            if phrase.skip_size != self.skip_size:
                searcher_size = f"{self.__class__.__name__} ({self.skip_size}"
                raise ValueError(f"phrase has different skip_size ({phrase.skip_size}) than {searcher_size}")
            self.phrases.add(phrase)
            if self.ignorecase:
                # print(f'indexing phrase {phrase.phrase_string} with lowercase')
                for skipgram in phrase.skipgrams_lower:
                    self.skipgram_index[skipgram.string].add(phrase)
                for skipgram_string in phrase.early_skipgram_index_lower:
                    # print('early skipgram_string:', skipgram_string)
                    self.early_skipgram_index[skipgram_string].add(phrase)
                for skipgram_string in phrase.late_skipgram_index_lower:
                    self.late_skipgram_index[skipgram_string].add(phrase)
            else:
                for skipgram in phrase.skipgrams:
                    self.skipgram_index[skipgram.string].add(phrase)
                for skipgram_string in phrase.early_skipgram_index:
                    self.early_skipgram_index[skipgram_string].add(phrase)
                for skipgram_string in phrase.late_skipgram_index:
                    self.late_skipgram_index[skipgram_string].add(phrase)
        if self.phrase_model is None:
            self.phrase_model = PhraseModel(phrases=list(self.phrases))

    def index_variants(self, variants: List[Union[str, Phrase]]) -> None:
        """Add a list of variant phrases to search for in texts.

        :param variants: a list of variants, either as string or as Phrase objects
        :type variants: List[Union[str, Phrase]]
        """
        for variant in variants:
            if isinstance(variant, str):
                variant = Phrase(variant, ngram_size=self.ngram_size, skip_size=self.skip_size)
            if variant.ngram_size != self.ngram_size:
                searcher_size = f"{self.__class__.__name__} ({self.ngram_size}"
                raise ValueError(f"variant has different ngram_size ({variant.ngram_size}) than {searcher_size}")
            if variant.skip_size != self.skip_size:
                searcher_size = f"{self.__class__.__name__} ({self.skip_size}"
                raise ValueError(f"variant has different skip_size ({variant.skip_size}) than {searcher_size}")
            self.variants.add(variant)
            if self.ignorecase:
                for skipgram in variant.skipgrams_lower:
                    self.variant_skipgram_index[skipgram.string].add(variant)
                for skipgram_string in variant.early_skipgram_index_lower:
                    self.variant_early_skipgram_index[skipgram_string].add(variant)
                for skipgram_string in variant.late_skipgram_index_lower:
                    self.variant_late_skipgram_index[skipgram_string].add(variant)
            else:
                for skipgram in variant.skipgrams:
                    self.variant_skipgram_index[skipgram.string].add(variant)
                for skipgram_string in variant.early_skipgram_index:
                    self.variant_early_skipgram_index[skipgram_string].add(variant)
                for skipgram_string in variant.late_skipgram_index:
                    self.variant_late_skipgram_index[skipgram_string].add(variant)

    def index_distractors(self, distractors: List[Union[str, Phrase]]) -> None:
        """Add a list of distractor phrases to filter out likely incorrect phrase matches.

        :param distractors: a list of distractors, either as string or as Phrase objects
        :type distractors: List[Union[str, Phrase]]
        """
        for distractor in distractors:
            if isinstance(distractor, str):
                distractor = Phrase(distractor, ngram_size=self.ngram_size, skip_size=self.skip_size)
            if distractor.ngram_size != self.ngram_size:
                searcher_size = f"{self.__class__.__name__} ({self.ngram_size}"
                raise ValueError(f"distractor has different ngram_size ({distractor.ngram_size}) than {searcher_size}")
            if distractor.skip_size != self.skip_size:
                searcher_size = f"{self.__class__.__name__} ({self.skip_size}"
                raise ValueError(f"distractor has different skip_size ({distractor.skip_size}) than {searcher_size}")
            self.distractors.add(distractor)
            if self.ignorecase:
                for skipgram in distractor.skipgrams_lower:
                    self.distractor_skipgram_index[skipgram.string].add(distractor)
                for skipgram_string in distractor.early_skipgram_index_lower:
                    self.distractor_early_skipgram_index[skipgram_string].add(distractor)
                for skipgram_string in distractor.late_skipgram_index_lower:
                    self.distractor_late_skipgram_index[skipgram_string].add(distractor)
            else:
                for skipgram in distractor.skipgrams:
                    self.distractor_skipgram_index[skipgram.string].add(distractor)
                for skipgram_string in distractor.early_skipgram_index:
                    self.distractor_early_skipgram_index[skipgram_string].add(distractor)
                for skipgram_string in distractor.late_skipgram_index:
                    self.distractor_late_skipgram_index[skipgram_string].add(distractor)

    def find_skipgram_matches(self, text: Dict[str, Union[str, int, float, list]],
                              include_variants: Union[None, bool] = None,
                              known_word_start_offset: Dict[int, Dict[str, any]] = None) -> SkipMatches:
        """Find all skipgram matches between text and phrases.

        :param text: the text object to match with phrases
        :type text: Dict[str, Union[str, int, float, list]]
        :param include_variants: boolean flag for whether to include phrase variants for finding matches
        :type include_variants: bool
        :param known_word_start_offset: a dictionary of known words and their text start_offsets based on exact matches
        :type known_word_start_offset: Dict[int, Dict[str, any]]
        :return: a SkipMatches object contain all skipgram matches
        :rtype: SkipMatches
        """
        # print(known_word_offset)
        # skipmatch_count = 0
        known_word = None
        if include_variants is None:
            include_variants = self.include_variants
        if known_word_start_offset is None:
            known_word_start_offset = {}
        # print(known_word_start_offset)
        skip_matches = SkipMatches(self.ngram_size, self.skip_size)
        text_string = text['text'].lower() if self.ignorecase else text['text']
        for skipgram in text2skipgrams(text_string, self.ngram_size, self.skip_size):
            # print(skipgram.start_offset, skipgram.string)
            # print("skipgram:", skipgram.string)
            if skipgram.start_offset in known_word_start_offset:
                known_word = known_word_start_offset[skipgram.start_offset]
                # print("known word start_offset reached:", known_word)
            if known_word and skipgram.start_offset == known_word["end"]:
                # print("end of known word start_offset reached:", known_word)
                known_word = None
            for phrase in self.skipgram_index[skipgram.string]:
                if phrase.max_start_offset > 0 and phrase.max_start_end < skipgram.start_offset + \
                        skipgram.length + self.max_length_variance:
                    # print(skipgram.offset, phrase.max_start_offset, phrase.max_start_end, phrase.phrase_string)
                    # print(f"skipping phrase {phrase.phrase_string} at offset", skipgram.offset)
                    continue
                if phrase.max_end_offset > 0 and phrase.max_end_end < skipgram.end_offset + \
                        skipgram.length + self.max_length_variance:
                    # print(skipgram.offset, phrase.max_end_offset, phrase.max_end_end, phrase.phrase_string)
                    # print(f"skipping phrase {phrase.phrase_string} at offset", skipgram.offset)
                    continue
                if known_word:
                    #if phrase.phrase_string not in self.phrase_model.word_in_phrase[known_word["word"]]:
                    #    print("skipping phrase because doesn't match known word:", phrase.phrase_string)
                    #    continue
                    if phrase.phrase_string in known_word["match_phrases"]:
                        # print("skipping phrase because found as exact match:", phrase.phrase_string)
                        continue
                # print("\tphrase has skip:", phrase.phrase_string)
                # skipmatch_count += 1
                # print("adding skipmatch", skipmatch_count)
                skip_matches.add_skip_match(skipgram, phrase)
            if include_variants:
                for phrase in self.variant_skipgram_index[skipgram.string]:
                    if known_word:
                        if phrase.phrase_string not in self.phrase_model.word_in_phrase[known_word["word"]]:
                            # print("skipping phrase because doesn't match known word:", phrase.phrase_string)
                            continue
                        if phrase.phrase_string in known_word["match_phrases"]:
                            # print("skipping phrase because found as exact match:", phrase.phrase_string)
                            continue
                    skip_matches.add_skip_match(skipgram, phrase)
        # print("final skipmatch count:", skipmatch_count)
        return skip_matches

    @staticmethod
    def filter_matches_by_offset_threshold(matches: List[PhraseMatch], debug: int = 0):
        filtered_matches = []
        for match in matches:
            if debug > 1:
                print('searcher.filter_matches_by_offset_threshold - match:\n\t', match.phrase.phrase_string,
                      match.phrase.max_start_offset, match.offset)
            if match.phrase.max_start_offset is None or match.phrase.max_start_offset == -1:
                if debug > 1:
                    print('no max start')
                filtered_matches.append(match)
            elif match.phrase.max_start_offset >= match.offset:
                if debug > 1:
                    print('lower than max start')
                filtered_matches.append(match)
            else:
                if debug > 1:
                    print('skipping')
                continue
        return filtered_matches



__version__ = '2.3.0'

from fuzzy_search.search.config import default_config
from fuzzy_search.search.phrase_searcher import FuzzyPhraseSearcher
from fuzzy_search.search.token_searcher import FuzzyTokenSearcher
from fuzzy_search.match.phrase_match import PhraseMatch
from fuzzy_search.phrase.phrase_model import PhraseModel


def make_searcher(phrases: any, config):
    phrase_model = PhraseModel(phrases, config)
    searcher = FuzzyPhraseSearcher(phrase_model=phrase_model, config=config)
    return searcher



import copy
from collections import defaultdict
from typing import Dict, List, Set, Union

from fuzzy_search.match.phrase_match import Candidate
from fuzzy_search.phrase.phrase import Phrase
from fuzzy_search.phrase.phrase_model import PhraseModel
from fuzzy_search.tokenization.string import SkipGram
from fuzzy_search.tokenization.string import score_levenshtein_similarity_ratio
from fuzzy_search.tokenization.token import Token


class SkipMatches:

    def __init__(self, ngram_size: int, skip_size: int):
        self.ngram_size = ngram_size
        self.skip_size = skip_size
        self.skip_length = ngram_size + skip_size
        self.match_set: Dict[Union[Phrase, Token, str], set] = defaultdict(set)
        self.match_start_offsets = defaultdict(list)
        self.match_end_offsets = defaultdict(list)
        self.match_skipgrams: Dict[Union[Phrase, Token, str], List[SkipGram]] = defaultdict(list)
        self.matches: Set[Union[Phrase, Token, str]] = set()

    def add_skip_match(self, skipgram: SkipGram, phrase: Union[Phrase, Token]) -> None:
        """Add a skipgram from a text that matches a phrase.

        :param skipgram: a skipgram from a text
        :type skipgram: SkipGram
        :param phrase: a phrase object that matches the skipgram
        :type phrase: Phrase
        """
        # track which skip grams of phrase are found in text
        self.match_set[phrase].add(skipgram.string)
        # track which text offsets match phrase skipgram
        self.match_start_offsets[phrase].append(skipgram.start_offset)
        self.match_end_offsets[phrase].append(skipgram.end_offset)
        self.match_skipgrams[phrase].append(skipgram)
        self.matches.add(phrase)


def get_skipset_overlap(phrase: Phrase, skip_matches: SkipMatches) -> float:
    """Calculate the overlap between the set of skipgrams of a text and the skipgrams of a phrase.

    :param phrase: a phrase object that has been matched against a text
    :type phrase: Phrase
    :param skip_matches: a SkipMatches object containing the skipgram matches between a text and a list of phrases
    :type skip_matches: SkipMatches
    :return: the fraction of skipgrams in the phrase that overlaps with the text
    :rtype: float
    """
    return len(skip_matches.match_set[phrase]) / len(phrase.skipgram_set)


def filter_skipgram_threshold(skip_matches: SkipMatches, skip_threshold: float) -> List[Phrase]:
    """Filter the skipgram matches based on the skipgram overlap threshold.

    :param skip_matches: the phrases that matches the text
    :type skip_matches: SkipMatches
    :param skip_threshold: the threshold for the skipgram overlap between a text and a phrase
    :type skip_threshold: float
    :return: the list of phrases with a skipgram overlap that meets the threshold
    :rtype: List[Phrase]
    """
    return [phrase for phrase in skip_matches.matches if get_skipset_overlap(phrase, skip_matches) >= skip_threshold]


def filter_overlapping_phrase_candidates(phrase_candidates: List[Candidate]) -> List[Candidate]:
    filtered: List[Candidate] = []
    if len(phrase_candidates) < 2:
        return phrase_candidates
    phrase_candidates.sort(key=lambda x: x.match_start_offset)
    prev_candidate = phrase_candidates[0]
    prev_score = score_levenshtein_similarity_ratio(prev_candidate.phrase.phrase_string, prev_candidate.match_string)
    for ci, curr_candidate in enumerate(phrase_candidates[1:]):
        if curr_candidate.match_end_offset > prev_candidate.match_start_offset:
            if curr_candidate.match_start_offset < prev_candidate.match_end_offset:
                # this candidate overlaps with the previous one, pick the best
                curr_score = score_levenshtein_similarity_ratio(curr_candidate.phrase.phrase_string,
                                                                curr_candidate.match_string)
                if curr_score > prev_score:
                    # this candidate is better, so skip the previous candidate
                    prev_candidate = curr_candidate
                    prev_score = curr_score
                elif curr_score == prev_score and len(curr_candidate.match_string) > len(prev_candidate.match_string):
                    # this candidate is longer with the same score, so skip the previous candidate
                    prev_candidate = curr_candidate
                    prev_score = curr_score
            else:
                # the previous candidate does not overlap with the current, so add it to filtered
                filtered.append(prev_candidate)
                prev_candidate = curr_candidate
                prev_score = score_levenshtein_similarity_ratio(curr_candidate.phrase.phrase_string,
                                                                curr_candidate.match_string)

    if len(filtered) == 0 or prev_candidate != filtered[-1]:
        filtered.append(prev_candidate)
    return filtered


def get_skipmatch_phrase_candidates(text: Dict[str, any], phrase: Phrase, skip_matches: SkipMatches,
                                    skipgram_threshold: float, max_length_variance: int = 1,
                                    ignorecase: bool = False, debug: int = 0) -> List[Candidate]:
    """Find all candidate matches for a given phrase and SkipMatches object.

    :param text: the text object to match with phrases
    :type text: Dict[str, any]
    :param phrase: a phrase to find candidate matches for
    :type phrase: Phrase
    :param skip_matches: a Skipmatches object with matches between a text and a list of phrases
    :type skip_matches: SkipMatches
    :param skipgram_threshold: a threshold for how many skipgrams should match between a phrase and a candidate
    :type skipgram_threshold: float
    :param max_length_variance: the maximum difference in length between candidate and phrase
    :type max_length_variance: int
    :param ignorecase: whether to ignore case when matching skip grams
    :type ignorecase: bool
    :param debug: level to show debug information
    :type debug: int
    :return: a list of candidate matches
    :rtype: List[Candidate]
    """
    candidates: List[Candidate] = []
    candidate = Candidate(phrase, max_length_variance=max_length_variance, ignorecase=ignorecase, debug=debug)
    last_index = len(skip_matches.match_start_offsets[phrase]) - 1
    if debug > 3:
        print(f"get_skipmatch_phrase_candidates - finding candidates for phrase "
              f"({len(phrase.phrase_string)}):", phrase.phrase_string)
        print('\t', skip_matches.match_start_offsets[phrase])
    for ci, curr_start_offset in enumerate(skip_matches.match_start_offsets[phrase]):
        next_start_offset = None if ci == last_index else skip_matches.match_start_offsets[phrase][ci + 1]
        if debug > 4:
            print('\t', ci, 'curr offset:', curr_start_offset, '\tskip:',
                  skip_matches.match_skipgrams[phrase][ci].string, '\tnext offset:', next_start_offset)
        # add current skipgram to the candidate
        candidate.add_skip_match(skip_matches.match_skipgrams[phrase][ci])
        if debug > 3 and abs(candidate.skip_match_length() - len(candidate.phrase.phrase_string)) < max_length_variance:
            skip = skip_matches.match_skipgrams[phrase][ci]
            print('\t', ci, curr_start_offset, "adding skip match:", skip.string, skip.start_offset, skip.length)
            print("\tcandidate skips:", [skip.string for skip in candidate.skipgram_list],
                  candidate.skip_match_length())
            print(candidate.get_skip_set_overlap(), candidate.get_match_string(text))
        # check if the current candidate is a potential match for the phrase
        if candidate.is_match(skipgram_threshold):
            candidate.match_string = candidate.get_match_string(text)
            if debug > 3:
                print("\tmeets threshold:", candidate.match_string)
            # if this candidate has enough skipgram overlap, yield it as a candidate match
            if len(candidates) == 0 or not candidate.same_candidate(candidates[-1]):
                candidates.append(copy.deepcopy(candidate))
            if candidate.shift_start_skip():
                # candidate string is longer than phrase string check if shifting the start creates
                # a better candidate and if so, add that as well
                candidate.match_string = candidate.get_match_string(text)
                candidates.append(copy.deepcopy(candidate))
        if next_start_offset and next_start_offset - curr_start_offset > skip_matches.ngram_size + skip_matches.skip_size + 1:
            # if the gap between the current skipgram and the next is larger than an entire skipgram
            # the next skipgram does not belong to this candidate
            # start a new candidate for the next skipgram
            if debug > 3:
                print('\tcurr_start_offset:', curr_start_offset, '\tnext_start_offset:', next_start_offset)
                print('\tstarting a new candidate')
            candidate = Candidate(phrase, max_length_variance=max_length_variance,
                                  ignorecase=ignorecase, debug=debug)
    # end of skipgrams reached, check if remaining candidate is a match
    if debug > 3:
        print('get_skipmatch_phrase_candidates - checking if final candidate is match:', candidate.is_match(skipgram_threshold))
    if candidate.is_match(skipgram_threshold):
        if len(candidates) == 0 or not candidate.same_candidate(candidates[-1]):
            candidate.match_string = candidate.get_match_string(text)
            candidates.append(copy.deepcopy(candidate))
        if candidate.shift_start_skip():
            # candidate string is longer than phrase string check if shifting the start creates
            # a better candidate and if so, add that as well
            candidate.match_string = candidate.get_match_string(text)
            candidates.append(copy.deepcopy(candidate))
    if debug > 2:
        print(f'get_skipmatch_phrase_candidates - returning {len(candidates)} candidates '
              f'for phrase "{phrase.phrase_string}"')
    if debug > 3:
        for candidate in candidates:
            print('\tCandidate:', candidate)
    return candidates


def get_skipmatch_candidates(text: Dict[str, any], skip_matches: SkipMatches,
                             skipgram_threshold: float, phrase_model: PhraseModel,
                             max_length_variance: int = 1, ignorecase: bool = False,
                             debug: int = 0) -> List[Candidate]:
    """Find all candidate matches for the phrases in a SkipMatches object.

    :param text: the text object to match with phrases
    :type text: Dict[str, any]
    :param skip_matches: a SkipMatches object with matches between a text and a list of phrases
    :type skip_matches: SkipMatches
    :param skipgram_threshold: a threshold for how many skipgrams should match between a phrase and a candidate
    :type skipgram_threshold: float
    :param phrase_model: a phrase model, either as dictionary or as PhraseModel object
    :type phrase_model: PhraseModel
    :param max_length_variance: the maximum difference in length between candidate and phrase
    :type max_length_variance: int
    :param ignorecase: whether to ignore case when matching skip grams
    :type ignorecase: bool
    :param debug: level to show debug information
    :type debug: int
    :return: a list of candidate matches
    :rtype: List[Candidate]
    """
    phrase_candidates = defaultdict(list)
    candidates: List[Candidate] = []
    for phrase in skip_matches.matches:
        if debug > 2:
            print("get_skipmatch_candidates - phrase:", phrase.phrase_string)
        if get_skipset_overlap(phrase, skip_matches) < skipgram_threshold:
            if debug > 2:
                print('get_skipmatch_candidates - below skipgram_threshold:', get_skipset_overlap(phrase, skip_matches))
            continue
        if phrase.phrase_string in phrase_model.is_variant_of:
            match_phrase = phrase_model.is_variant_of[phrase.phrase_string]
        else:
            match_phrase = phrase.phrase_string
        phrase_candidates[match_phrase] += get_skipmatch_phrase_candidates(text, phrase, skip_matches,
                                                                           skipgram_threshold,
                                                                           max_length_variance=max_length_variance,
                                                                           ignorecase=ignorecase, debug=debug)
    if debug > 1:
        print(f"get_skipmatch_candidates - phrases with candidate matches:")
        for phrase_string in phrase_candidates:
            if len(phrase_candidates[phrase_string]) > 0:
                print(f"\tphrase_string: {phrase_string}\tnum candidates: {len(phrase_candidates[phrase_string])}")

    if debug > 1:
        print(f"get_skipmatch_candidates - phrases with overlap filtered candidates:")
    for phrase_string in phrase_candidates:
        filtered_candidates = filter_overlapping_phrase_candidates(phrase_candidates[phrase_string])
        if debug > 1 and len(filtered_candidates) > 0:
            print(f"\tphrase_string: {phrase_string}\tnum overlap filtered candidates: {len(filtered_candidates)}")
            if debug > 2:
                for candidate in filtered_candidates:
                    print('\t', candidate.match_string, candidate.match_start_offset, candidate.match_end_offset)
        candidates += filtered_candidates
    if debug > 0:
        print(f'get_skipmatch_candidates - returning {len(candidates)} candidates')
    return candidates

